#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for quantum gates.
"""

import numpy

__all__ = [
    "Gate"
]


class Gate:
    r"""Class to obtain the matrix of a quantum gate.
    """

    @classmethod
    def I(cls) -> numpy.ndarray:
        r"""Return the matrix of identity gate.

        Returns:
            numpy.ndarray: identity gate
        """
        return numpy.array([[1, 0], [0, 1]], dtype=complex)

    @classmethod
    def X(cls) -> numpy.ndarray:
        r"""Return the matrix of Pauli-X gate.

        Returns:
            numpy.ndarray: Pauli-X gate
        """
        return numpy.array([[0, 1], [1, 0]], dtype=complex)

    @classmethod
    def Y(cls) -> numpy.ndarray:
        r"""Return the matrix of Pauli-Y gate.

        Returns:
            numpy.ndarray: Pauli-Y gate
        """
        return numpy.array([[0, -1j], [1j, 0]], dtype=complex)

    @classmethod
    def Z(cls) -> numpy.ndarray:
        r"""Return the matrix of Pauli-Z gate.

        Returns:
            numpy.ndarray: Pauli-Z gate
        """
        return numpy.array([[1, 0], [0, -1]], dtype=complex)

    @classmethod
    def H(cls) -> numpy.ndarray:
        r"""Return the matrix of Hadamard gate.

        Returns:
            numpy.ndarray: Hadamard gate
        """
        return numpy.array([[1, 1], [1, -1]], dtype=complex) / numpy.sqrt(2.0)

    @classmethod
    def S(cls) -> numpy.ndarray:
        r"""Return the matrix of phase gate.

        Returns:
            numpy.ndarray: phase gate
        """
        return numpy.array([[1, 0], [0, 1j]], dtype=complex)

    @classmethod
    def T(cls) -> numpy.ndarray:
        r"""Return the matrix of T gate.

        Returns:
            numpy.ndarray: T gate
        """
        return numpy.array([[1, 0], [0, numpy.exp(1j * numpy.pi / 4)]], dtype=complex)

    @classmethod
    def Rx(cls, theta: float) -> numpy.ndarray:
        r"""Return the matrix of Rx gate.

        Returns:
            numpy.ndarray: Rx gate
        """
        return numpy.array([[numpy.cos(theta / 2), -1j * numpy.sin(theta / 2)],
                            [-1j * numpy.sin(theta / 2), numpy.cos(theta / 2)]], dtype=complex)

    @classmethod
    def Ry(cls, theta: float) -> numpy.ndarray:
        r"""Return the matrix of Ry gate.

        Returns:
            numpy.ndarray: Ry gate
        """
        return numpy.array([[numpy.cos(theta / 2), - numpy.sin(theta / 2)],
                            [numpy.sin(theta / 2), numpy.cos(theta / 2)]], dtype=complex)

    @classmethod
    def Rz(cls, theta: float) -> numpy.ndarray:
        r"""Return the matrix of Rz gate.

        Returns:
            numpy.ndarray: Rz gate
        """
        return numpy.array([[numpy.exp(- 1j * theta / 2), 0],
                            [0, numpy.exp(1j * theta / 2)]], dtype=complex)

    @classmethod
    def U(cls, theta: float, phi: float, gamma: float) -> numpy.ndarray:
        r"""Return the matrix of U gate.

        .. math::

            U (\theta, \phi, \gamma) = Rz(\phi) Ry(\theta) Rz(\gamma)

        This is usually used in the measurement-based quantum computation.

        Args:
            theta (float): Ry rotation angle
            phi (float): left Rz rotation angle
            gamma (float): right Rz rotation angle

        Returns:
            numpy.ndarray: U gate
        """
        return Gate.Rz(phi) @ Gate.Ry(theta) @ Gate.Rz(gamma)

    @classmethod
    def U3(cls, theta: float, phi: float, gamma: float) -> numpy.ndarray:
        r"""Return the matrix of U3 gate.

        .. math::

            U3 (\theta, \phi, \gamma) = Rz(\phi) Rx(\theta) Rz(\gamma)

        Args:
            theta (float): Rx rotation angle
            phi (float): left Rz rotation angle
            gamma (float): right Rz rotation angle

        Returns:
            numpy.ndarray: U3 gate
        """
        return Gate.Rz(phi) @ Gate.Rx(theta) @ Gate.Rz(gamma)

    @classmethod
    def CZ(cls) -> numpy.ndarray:
        r"""Return the matrix of control-Z gate.

        Returns:
            numpy.ndarray: control-Z gate
        """
        return numpy.array([[1, 0, 0, 0],
                            [0, 1, 0, 0],
                            [0, 0, 1, 0],
                            [0, 0, 0, -1]], dtype=complex)

    @classmethod
    def CNOT(cls) -> numpy.ndarray:
        r"""Return the matrix of control-NOT gate.

        Returns:
            numpy.ndarray: control-CNOT gate
        """
        return numpy.array([[1, 0, 0, 0],
                            [0, 1, 0, 0],
                            [0, 0, 0, 1],
                            [0, 0, 1, 0]], dtype=complex)

    @classmethod
    def SWAP(cls) -> numpy.ndarray:
        r"""Return the matrix of SWAP gate.

        Returns:
            numpy.ndarray: SWAP gate
        """
        return numpy.array([[1, 0, 0, 0],
                            [0, 0, 1, 0],
                            [0, 1, 0, 0],
                            [0, 0, 0, 1]], dtype=complex)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for protocols.
"""

from abc import ABC
from typing import List, Union
import networkx as nx
import matplotlib.pyplot as plt
from qcompute_qnet.core.des import Event, Scheduler, EventHandler

__all__ = [
    "Protocol",
    "SubProtocol",
    "ProtocolStack"
]


class Protocol(ABC):
    r"""Abstract protocol class for designing a protocol.

    Attributes:
        name (str): name of the protocol
        owner (ProtocolStack): owner of the protocol
        node (Node): node where the protocol stack is loaded
        scheduler (Scheduler): event scheduler of the protocol
        agenda (List): events scheduled for the protocol
        signed_events (List): events scheduled by the protocol
    """

    def __init__(self, name=None):
        r"""Constructor for Protocol class.

        Args:
            name (str): name of the protocol
        """

        self.name = 'Protocol' if name is None else name
        self.owner = self
        self.node = None
        self.scheduler = Scheduler(owner=self)
        self.agenda = []
        self.signed_events = []

    @property
    def is_top(self) -> bool:
        r"""Check if the protocol is at the top of the protocol stack.

        Returns:
            bool: whether the protocol is a top protocol
        """
        return len(self.upper_protocols) == 0

    @property
    def is_bottom(self) -> bool:
        r"""Check if the protocol is at the bottom of the protocol stack.

        Returns:
            bool: whether the protocol is a bottom protocol
        """
        return len(self.lower_protocols) == 0

    @property
    def upper_protocols(self) -> List["Protocol"]:
        r"""Return the upper protocols of the current protocol.

        Returns:
            List: a list of upper protocols
        """
        assert isinstance(self.owner, ProtocolStack), f"Not attached to any protocol stack yet!"
        return list(self.owner.stack.predecessors(self))

    @property
    def lower_protocols(self) -> List["Protocol"]:
        r"""Return the lower protocols of the current protocol.

        Returns:
            List: a list of lower protocols
        """
        assert isinstance(self.owner, ProtocolStack), f"Not attached to any protocol stack yet!"
        return list(self.owner.stack.successors(self))

    def start(self, **kwargs) -> None:
        r"""Start the protocol.

        This should be overridden in a specific protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        pass

    def send_upper(self, upper_protocol: type, **kwargs) -> None:
        r"""Send a message to upper protocols of a specified type.

        Args:
            upper_protocol (type): type of the upper protocols to receive the message
            **kwargs: keyword arguments of the message
        """
        assert not self.is_top, f"'{self.name}' has no upper protocols."

        for proto in self.upper_protocols:
            if isinstance(proto, upper_protocol):
                proto.receive_lower(type(self), **kwargs)

    def send_lower(self, lower_protocol: type, **kwargs) -> None:
        r"""Send a message to lower protocols of a specified type.

        Args:
            lower_protocol (type): type of the lower protocols to receive the message
            **kwargs: keyword arguments of the message
        """
        assert not self.is_bottom, f"'{self.name}' has no lower protocols."

        for proto in self.lower_protocols:
            if isinstance(proto, lower_protocol):
                proto.receive_upper(type(self), **kwargs)

    def receive_upper(self, upper_protocol: type, **kwargs) -> None:
        r"""Receive a message from an upper protocol.

        Args:
            upper_protocol (type): type of the upper protocol
            **kwargs: keyword arguments of the message
        """
        pass

    def receive_lower(self, lower_protocol: type, **kwargs) -> None:
        r"""Receive a message from a lower protocol.

        Args:
            lower_protocol (type): type of the lower protocol
            **kwargs: keyword arguments of the message
        """
        pass

    def receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receive a classical message from the node.

        Args:
            msg (ClassicalMessage): received classical message
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        pass

    def receive_quantum_msg(self, msg: "QuantumMessage", **kwargs) -> None:
        r"""Receive a quantum message from the node.

        Args:
            msg (QuantumMessage): received quantum message
            **kwargs (Any): keyword arguments for receiving the quantum message
        """
        pass

    def print_agenda(self) -> None:
        r"""Print the events scheduled for the protocol.
        """
        df = Event.events_to_dataframe(self.agenda)
        print(f"\nAgenda of {self.name} (unsorted):\n{df.to_string()}")

    def print_signed_events(self) -> None:
        r"""Print the events scheduled by the protocol.
        """
        df = Event.events_to_dataframe(self.signed_events)
        print(f"\nSigned events by {self.name} (unsorted):\n{df.to_string()}")


class SubProtocol:
    r"""Class for the sub-protocols built in a protocol.

    Note:
        A sub-protocol defines the behaviors of a specific role in a protocol.
        Taking teleportation protocol as an example, we can define sub-protocols for sender and receiver respectively.
        Then these sub-protocols together forms a full protocol for teleportation.

    Attributes:
        super_protocol (Protocol): protocol in which the sub-protocol is built
    """

    def __init__(self, super_protocol: Protocol):
        r"""Constructor for SubProtocol class.

        Args:
            super_protocol (Protocol): protocol in which the sub-protocol is built
        """
        self.super_protocol = super_protocol

    @property
    def node(self) -> "Node":
        r"""Get the node of its super protocol.

        Returns:
            Node: node of super protocol
        """
        return self.super_protocol.node


class ProtocolStack:
    r"""Class for building a protocol stack.

    Note that the stack relation is maintained as a directed graph.

    Attributes:
        name (str): name of the ProtocolStack
        owner (Node): owner of the ProtocolStack
        stack (networkx.DiGraph): stack relation of protocols
        config (Dict[str, Dict]): configurations of the protocols
    """

    def __init__(self, name=None, protocol=None):
        r"""Constructor for ProtocolStack class.

        Args:
            name (str): name of the ProtocolStack
        """

        self.name = 'Protocol Stack' if name is None else name
        self.owner = self
        self.stack = nx.DiGraph()
        self.config = {}
        if protocol is not None:
            self.build(protocol)

    @property
    def protocols(self) -> List["Protocol"]:
        r"""Return a list of all protocols in the protocol stack.

        Returns:
            List[Protocol]: protocols in the protocol stack
        """
        return list(self.stack.nodes)

    def get_protocol(self, proto_type: type) -> "Protocol":
        r"""Get a protocol from the protocol stack by its type.

        Args:
            proto_type (type): type of the protocol

        Returns:
            Protocol: protocol of the given type
        """
        assert any(isinstance(proto, proto_type) for proto in self.protocols), "Cannot find a protocol with the given type."
        for proto in self.protocols:
            if isinstance(proto, proto_type):
                return proto

    def set(self, protocol_type: str, **kwargs) -> None:
        r"""Set configuration of a specified protocol type by given parameters.

        Args:
            protocol_type (str): type of the protocol to set
            **kwargs: keyword arguments to set
        """
        assert isinstance(protocol_type, str), "'protocol' should be a str."

        self.config[protocol_type] = kwargs

    def build(self, relation: Union["Protocol", List["Protocol"]]) -> None:
        r"""Build a protocol stack from the protocol relation.

        Args:
            relation (Union[Protocol, List[Protocol]]): a single protocol or a list of protocol relations
        """
        if isinstance(relation, Protocol):
            self.stack.add_node(relation)
            relation.owner = self
        elif isinstance(relation, list):
            self.__update_from_relations(relation)
        else:
            raise TypeError("Should input a single protocol or a list of protocol relations!")
        from qcompute_qnet.topology.node import Node
        if isinstance(self.owner, Node):
            self.sync_env(self.owner)

    def __update_from_relations(self, relation: list) -> None:
        r"""Update the protocol stack from the protocol relation.

        Args:
            relation (list): a list of protocol relations
        """
        assert all(edge[0] != edge[1] for edge in relation), "Invalid relation input."

        # Build or update the relation graph from edges, nodes will be added automatically
        self.stack.add_edges_from(relation)
        # Update protocol owners
        for protocol in self.protocols:
            protocol.owner = self

    def sync_env(self, node: "Node") -> None:
        r"""Synchronize the simulation environment for each protocol.

        Args:
            node (Node): node that the protocol stack is loaded to
        """
        for protocol in self.protocols:
            protocol.node = node
            protocol.scheduler.env = node.env

    def update(self, protocol: "Protocol", upper_protocols=None, lower_protocols=None) -> None:
        r"""Add a protocol to the protocol stack.

        Args:
            protocol (Protocol): protocol to add
            upper_protocols (List[Protocol]): upper protocols of the protocol
            lower_protocols (List[Protocol]): lower protocols of the protocol
        """
        assert upper_protocols is not None or lower_protocols is not None,\
            f"Should assign at least an upper protocol or a lower protocol!"

        from qcompute_qnet.topology.node import Node
        if isinstance(self.owner, Node):
            protocol.node = self.owner
            protocol.scheduler.env = self.owner.env
        relation = []
        if upper_protocols is not None:
            for upper in upper_protocols:
                relation.append((upper, protocol))
        if lower_protocols is not None:
            for lower in lower_protocols:
                relation.append((protocol, lower))
        self.__update_from_relations(relation)

    def print_stack(self) -> None:
        r"""Print the stack relation as a graph.
        """
        labels = {proto: proto.name for proto in self.protocols}
        pos = nx.spring_layout(self.stack)
        nx.draw(self.stack, pos=pos, labels=labels, with_labels=True, edge_color='r', arrowsize=20)
        plt.show()

    def start(self, **kwargs) -> None:
        r"""Start the protocol stack.

        Args:
            **kwargs: keyword arguments to start the protocol stack
        """
        assert self.owner is not None, f"Should load the protocol stack to a node first!"

        for proto in self.protocols:
            if proto.is_top:
                handler = EventHandler(proto, "start", **kwargs)
                proto.scheduler.schedule_now(handler)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for error correction protocols.
"""

from abc import ABC
from qcompute_qnet.protocols.protocol import Protocol

__all__ = [
    "ErrorCorrection",
    "Cascade",
    "LDPC"
]


class ErrorCorrection(Protocol, ABC):
    r"""Class for error correction protocols.
    """

    def __init__(self, name: str):
        r"""Constructor for ErrorCorrection class.

        Args:
            name (str): name of the error correction protocol
        """
        super().__init__(name)


class Cascade(ErrorCorrection):
    r"""Class for the cascade protocol.
    """

    def __init__(self, name: str):
        r"""Constructor for Cascade class.

        Args:
            name (str): name of the cascade protocol
        """
        super().__init__(name)


class LDPC(ErrorCorrection):
    r"""Class for the low-density parity-check protocol.
    """

    def __init__(self, name: str):
        r"""Constructor for LDPC class.

        Args:
            name (str): name of the LDPC protocol
        """
        super().__init__(name)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for key generation protocols.
"""

from abc import ABC, abstractmethod
from enum import Enum, unique
from typing import List, Dict, Tuple
import numpy as np
from qcompute_qnet.core.des import EventHandler
from qcompute_qnet.protocols.protocol import Protocol
from qcompute_qnet.messages.message import ClassicalMessage
from qcompute_qnet.quantum.basis import Basis

__all__ = [
    "KeyGeneration",
    "PrepareAndMeasure",
    "BB84",
    "DecoyBB84"
]


class KeyGeneration(Protocol, ABC):
    r"""Class for key generation protocols.

    Attributes:
        peer (QKDNode): peer node to implement the same task
        start_time (int): start time of the key generation protocol
        key_num (int): number of the sifted keys
        key_length (int): length of the sifted keys
        status (Status): the status of the key generation protocol
        sifted_keys (list): list of the sifted keys
        key_rate (float): rate of key generation
        error_rate (float): error rate of key generation
    """

    def __init__(self, name: str, peer=None):
        r"""Constructor for KeyGeneration class.

        Args:
            name (str): name of the key generation protocol
            peer (QKDNode): peer node to implement the same task
        """
        super().__init__(name)
        self.peer = peer
        self.start_time = 0
        self.key_num = float("inf")
        self.key_length = 256
        self.status = KeyGeneration.Status.READY
        self.sifted_keys = []
        self.key_rate = 0
        self.error_rate = 0

    @unique
    class Status(Enum):
        r"""Class for the status of the key generation protocol.
        """

        READY = "Ready"
        WORKING = "Working"
        DONE = "Done"

    def receive_upper(self, upper_protocol: type, **kwargs) -> None:
        r"""Receive a message from an upper protocol.

        Args:
            upper_protocol (type): upper protocol that sends the message
            **kwargs: received keyword arguments
        """
        if kwargs['peer'] == self.peer:
            self.start(**kwargs)

    def finish(self) -> None:
        r"""Finish the key generation protocol.
        """
        self.node.env.logger.info(f"{self.node.name} finished key generation with {self.peer.name}")
        self.node.env.logger.info(f"{self.node.name}'s sifted keys (with {self.peer.name} in decimal): "
                                  f"{self.sifted_keys}")

        self.status = KeyGeneration.Status.DONE

        if self.is_top:
            self.node.keys[self.peer] = self.sifted_keys
        else:
            # TODO: only consider QKDRouting as an upper protocol in this version
            qkd_routing = type(self.upper_protocols[-1])
            towards = 'up' if self.role == PrepareAndMeasure.Role.TRANSMITTER else 'down'

            handler = EventHandler(self, "send_upper", [qkd_routing],
                                   peer=self.peer, sifted_keys=self.sifted_keys, towards=towards)
            self.node.scheduler.schedule_now(handler)


class PrepareAndMeasure(KeyGeneration, ABC):
    r"""Class for the prepare-and-measure type of quantum key distribution protocols.

    Attributes:
        role (Role): role of the prepare-and-measure protocol
        round (int): round of photon emission
        pulse_num (int): number of photon pulses
        light_duration (int): duration of photon emissions in picoseconds
        emission_rest (int): gap between photon emissions
        tx_bases_ratio (list): ratio of the preparation bases
        tx_bases_list (list): preparation bases for the transmitter
        tx_bits_list (list): chosen bits for the transmitter
        tx_key_bits (list): key bits for the transmitter
        rx_bases_ratio (list): ratio of the measurement bases
        rx_bases_list (list): measurement bases for the receiver
        rx_bits_list (list): chosen bits for the receiver
        rx_key_bits (list): key bits for the receiver
    """

    def __init__(self, name: str, peer=None):
        r"""Constructor for PrepareAndMeasure class.

        Args:
            name (str): name of the prepare-and-measure protocol
            peer (QKDNode): peer node to implement the same task
        """
        super().__init__(name, peer)
        self.role = None
        self.round = 0
        self.pulse_num = 0
        self.light_duration = 0
        self.emission_rest = 0

        self.tx_bases_ratio = None
        self.tx_bases_list = None
        self.tx_bits_list = None
        self.tx_key_bits = None

        self.rx_bases_ratio = None
        self.rx_bases_list = None
        self.rx_bits_list = None
        self.rx_key_bits = None

    @unique
    class Role(Enum):
        r"""Class for the role of the prepare-and-measure protocol.
        """

        TRANSMITTER = "Transmitter"
        RECEIVER = "Receiver"

    def set(self, **kwargs) -> None:
        r"""Set given parameters.

        Args:
            **kwargs: keyword arguments to set
        """
        for attr in kwargs:
            if attr == "tx_bases_ratio":
                assert isinstance(kwargs[attr], list), "'tx_bases_ratio' should be a list."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "rx_bases_ratio":
                assert isinstance(kwargs[attr], list), "'rx_bases_ratio' should be a list."
                self.__setattr__(attr, kwargs[attr])
            else:
                raise TypeError(f"Setting {attr} is not allowed in {self.name}")

    def receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receive a classical message from the node.

        Args:
            msg (ClassicalMessage): classical message used for the prepare-and-measure protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        # Check if the message is for self and the protocol is working
        if msg.src == self.peer and self.status == KeyGeneration.Status.WORKING:
            # Different roles lead to different operations
            if self.role == PrepareAndMeasure.Role.TRANSMITTER:
                self.transmitter_receive_classical_msg(msg)
            elif self.role == PrepareAndMeasure.Role.RECEIVER:
                self.receiver_receive_classical_msg(msg)

    def receive_quantum_msg(self, msg: "QuantumMessage", **kwargs) -> None:
        r"""Receive a quantum message from the node.

        Args:
            msg (Message): received quantum message
            **kwargs (Any): keyword arguments for receiving the quantum message
        """
        if self.role == PrepareAndMeasure.Role.RECEIVER:
            if isinstance(self.node.polar_detector, dict):
                src = kwargs['src']
                self.node.polar_detector[src].receive(msg)
            else:
                self.node.polar_detector.receive(msg)

    @abstractmethod
    def transmitter_receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Transmitter receives a classical message from the receiver.

        Args:
            msg (ClassicalMessage): classical message for the prepare-and-measure protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        pass

    @abstractmethod
    def receiver_receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receiver receives a classical message from the transmitter.

        Args:
            msg (ClassicalMessage): classical message for the prepare-and-measure protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        pass

    def receiver_get_bits(self, **kwargs) -> List:
        r"""Get bits from the detectors' time records.

        Warnings:
            We regard the detection events as invalid if both detectors click.
            This may not consist with some literatures where a random choice is made if both detectors click.
            However, this makes no much difference to the key rate's estimation.
            Users can override this method if necessary.

        Args:
            **kwargs: keyword argument to specify the polar detector for different nodes

        Returns:
            list: receiver's bits list
        """
        bits = [None] * self.pulse_num
        if isinstance(self.node.polar_detector, dict):
            peer = kwargs['src']
            records_zero = self.node.polar_detector[peer].detectors[0].pop_records()
            records_one = self.node.polar_detector[peer].detectors[1].pop_records()
        else:
            records_zero = self.node.polar_detector.detectors[0].pop_records()
            records_one = self.node.polar_detector.detectors[1].pop_records()

        for record in records_zero:
            index = record[0]
            bits[index] = 0

        for record in records_one:
            index = record[0]
            bits[index] = None if bits[index] == 0 else 1

        return bits

    def key_rate_estimation(self) -> float:
        r"""Estimate sifted key rate.

        Returns:
            float: estimated sifted key rate
        """
        self.key_rate = ((len(self.sifted_keys) * self.key_length + len(self.tx_key_bits)) / 1000) / \
                        ((self.node.env.now - self.start_time) * 1e-12)
        self.node.env.logger.info(f"Sifted key rate: {self.key_rate:.4f} kbit/s")
        return self.key_rate

    def error_rate_estimation(self) -> float:
        r"""Estimate error rate of the sifted keys.

        Returns:
            float: estimated error rate of the sifted keys
        """
        # Find the peer protocol
        for proto in self.peer.protocol_stack.protocols:
            if isinstance(proto, KeyGeneration) and proto.peer == self.node and proto.role != self.role:
                peer_proto = proto

        # Calculate error rate of all sifted keys
        num_errors = 0
        for i in range(self.key_num):
            key_xor = self.sifted_keys[i] ^ peer_proto.sifted_keys[i]  # in decimal
            num_errors += bin(key_xor).count('1')  # count the number of errors
        self.error_rate = num_errors / (self.key_num * self.key_length)
        self.node.env.logger.info(f"Key error rate: {self.error_rate:.4f}")
        return self.error_rate

    def statistics_estimation(self) -> Tuple[float, float]:
        r"""Estimate key rate and error rate of the sifted keys.

        Returns:
            Tuple[float, float]: estimated key rate and error rate
        """
        key_rate = self.key_rate_estimation()
        error_rate = self.error_rate_estimation()
        return key_rate, error_rate


class BB84(PrepareAndMeasure):
    r"""Class for the BB84 key generation protocol.

    Attributes:
        basis_encoding (dict): map an int to a measurement basis
    """

    def __init__(self, name: str, peer=None):
        r"""Constructor for BB84 class.

        Args:
            name (str): name of the BB84 protocol
            peer (QKDNode): peer node to implement the same task
        """
        super().__init__(name, peer)
        self.basis_encoding = {0: Basis.Z(), 1: Basis.X()}

    class Message(ClassicalMessage):
        r"""Class for the classical control messages in BB84 protocol.
        """

        def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
            r"""Constructor for Message class.

            Args:
                src (Node): source of Message
                dst (Node): destination of Message
                protocol (type): protocol of Message
                data (Dict): message content
            """
            super().__init__(src, dst, protocol, data)

        @unique
        class Type(Enum):
            r"""Class for Message types.
            """

            START = "Start"
            BASES_REQUEST = "Bases request"
            BASES = "Bases"
            MATCHING = "Matching"

    def start(self, **kwargs) -> None:
        r"""Start the BB84 protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        assert self.node is not None and self.peer is not None, f"Should load the protocol stack to a node first!"

        # Initial settings
        self.role = kwargs['role']
        if "key_num" in kwargs.keys():
            self.key_num = kwargs['key_num']
        if "key_length" in kwargs.keys():
            self.key_length = kwargs['key_length']
        self.status = KeyGeneration.Status.WORKING
        self.sifted_keys = []
        self.round = 0

        if self.role == PrepareAndMeasure.Role.TRANSMITTER:
            self.start_time = self.node.env.now  # record the start time of BB84 protocol
            if self.tx_bases_ratio is None:
                self.tx_bases_ratio = [0.5, 0.5]
            self.tx_bases_list = []
            self.tx_bits_list = []
            self.tx_key_bits = []
            self.transmitter_start_protocol()

        elif self.role == PrepareAndMeasure.Role.RECEIVER:
            if self.rx_bases_ratio is None:
                self.rx_bases_ratio = [0.5, 0.5]
            self.rx_bases_list = []
            self.rx_bits_list = []
            self.rx_key_bits = []

    def transmitter_start_protocol(self) -> None:
        r"""Transmitter starts the BB84 protocol.

        Transmitter sends a classical message to the receiver for device synchronization
        and prepares to emit photon pulses.
        """
        self.node.env.logger.info(f"{self.node.name} started BB84 protocol with {self.peer.name}")
        # Select the corresponding photon source for key generation
        photon_source = self.node.photon_source[self.peer] if isinstance(self.node.photon_source, dict)\
            else self.node.photon_source

        self.pulse_num = round(self.key_length / photon_source.mean_photon_num)
        self.light_duration = round(self.pulse_num / photon_source.frequency * 1e12)
        self.emission_rest = round(1e12 / photon_source.frequency)

        transmitter_emit_time = self.node.env.now + self.node.cchannel(self.peer).delay
        clock = transmitter_emit_time + self.node.qchannel(self.peer).delay  # clock time for photon reception
        start_msg = BB84.Message(src=self.node,
                                 dst=self.peer,
                                 protocol=BB84,
                                 data={'type': BB84.Message.Type.START,
                                       'clock': clock,
                                       'source_frequency': photon_source.frequency,
                                       'light_duration': self.light_duration,
                                       'emission_rest': self.emission_rest})
        self.node.send_classical_msg(dst=self.peer, msg=start_msg)

        # Transmitter begins photon emission as soon as the receiver receives the message
        handler = EventHandler(self, "transmitter_start_emission")
        self.scheduler.schedule_after(self.node.cchannel(self.peer).delay, handler)

    def transmitter_start_emission(self) -> None:
        r"""Transmitter starts to emit photon pulses.

        Transmitter randomly generates bits and encoding bases for photon emission.
        """
        if self.status == KeyGeneration.Status.WORKING:
            self.round += 1
            self.node.env.logger.debug(f"{self.node.name} began photon emission for BB84 with {self.peer.name}"
                                       f" (Round {self.round})")

            # Generate transmitter's random bases and bits
            bases = list(np.random.choice([0, 1], self.pulse_num, p=self.tx_bases_ratio))
            bits = list(np.random.choice([0, 1], self.pulse_num))  # uniform distribution
            self.tx_bases_list.append(bases)
            self.tx_bits_list.append(bits)

            bases = [self.basis_encoding[basis] for basis in bases]
            states = []
            for i, basis in enumerate(bases):
                # Prepare states |0>, |1>, |+>, |-> (in the form of density matrices) for photon emission
                states.append(basis[bits[i]] @ np.conjugate(basis[bits[i]]).T)

            photon_source = self.node.photon_source[self.peer] if isinstance(self.node.photon_source, dict) \
                else self.node.photon_source

            handler = EventHandler(photon_source, "emit", [self.peer, states])
            self.scheduler.schedule_now(handler)

            # Schedule next round of photon emission
            handler = EventHandler(self, "transmitter_start_emission")
            # Rest for a while to avoid scheduling the event during cool-down period of the photon source
            self.scheduler.schedule_after(self.light_duration + self.emission_rest, handler)

    def receiver_stop_reception(self) -> None:
        r"""Receiver stops photon reception.

        Receiver pops the measurement outcomes and resets the measurement bases to
        prepare for receiving photon pulses in the next round.
        """
        # Select the corresponding detector for key generation
        polar_detector = self.node.polar_detector[self.peer] if isinstance(self.node.polar_detector, dict) \
            else self.node.polar_detector

        if self.status == KeyGeneration.Status.WORKING:
            self.round += 1
            self.node.env.logger.debug(f"{self.node.name} finished receiving photons from {self.peer.name}"
                                       f" (Round {self.round})")

            self.rx_bits_list.append(self.receiver_get_bits(src=self.peer))
            # Reset measurement bases for the next round of photon reception
            new_bases = list(np.random.choice([0, 1], self.pulse_num, p=self.rx_bases_ratio))
            self.rx_bases_list.append(new_bases)
            clock = self.node.env.now + self.emission_rest  # set the next clock time

            new_bases = [self.basis_encoding[basis] for basis in new_bases]
            # Reset parameters for polarization detector
            polar_detector.set_beamsplitter(clock=clock, bases=new_bases)

            # Schedule next round of stopping photon reception
            handler = EventHandler(self, "receiver_stop_reception")
            self.scheduler.schedule_at(clock + self.light_duration, handler)

            # Ask the transmitter to announce the bases for key generation
            request_bases_msg = BB84.Message(src=self.node,
                                             dst=self.peer,
                                             protocol=BB84,
                                             data={'type': BB84.Message.Type.BASES_REQUEST,
                                                   'round': self.round})
            self.node.send_classical_msg(dst=self.peer, msg=request_bases_msg)

        elif self.status == KeyGeneration.Status.DONE:
            polar_detector.detectors[0].turn_off()
            polar_detector.detectors[1].turn_off()

    def transmitter_receive_classical_msg(self, msg: "BB84.Message", **kwargs) -> None:
        r"""Transmitter receives a BB84 message from the receiver.

        Args:
            msg (BB84.Message): message for BB84 protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        # Filter the message that is not intended for this instance
        if msg.data['type'].name not in ['MATCHING', 'BASES_REQUEST']:
            return

        photon_round = msg.data.get('round')

        # Check photon round to filter the outdated messages resulting from BB84 protocol's reusing
        if photon_round is not None and self.round < photon_round:
            self.node.env.logger.debug(f"Message with round {photon_round} was discarded at local round {self.round}")
            return

        if msg.data['type'] == BB84.Message.Type.BASES_REQUEST:
            self.node.env.logger.debug(f"{self.node.name} received bases request from"
                                       f" {self.peer.name} (Round {photon_round})")

            # Pop the earliest bases and send to the receiver
            bases = self.tx_bases_list.pop(0)

            bases_msg = BB84.Message(src=self.node,
                                     dst=self.peer,
                                     protocol=BB84,
                                     data={'type': BB84.Message.Type.BASES,
                                           'peer_bases': bases,
                                           'round': photon_round})
            self.node.send_classical_msg(dst=self.peer, msg=bases_msg)

        elif msg.data['type'] == BB84.Message.Type.MATCHING:
            self.node.env.logger.debug(f"{self.node.name} received matching indices from {self.peer.name}"
                                       f" (Round {photon_round})")

            matching_list = msg.data['matching_list']
            bits = self.tx_bits_list.pop(0)  # pop the earliest bits for index comparison
            for i in matching_list:
                self.tx_key_bits.append(bits[i])

            # Transmitter generates a valid key successfully
            if len(self.tx_key_bits) >= self.key_length:
                self.node.env.logger.info(f"{self.node.name} generated a valid key with {self.peer.name}"
                                          f" (Round {photon_round})")
                self.node.env.logger.info(f"{self.node.name}'s sifted key: {self.tx_key_bits[0: self.key_length]}")

                # Store the generated key as a decimal integer
                transmitter_key = int("".join(str(self.tx_key_bits.pop(0)) for _ in range(self.key_length)), base=2)
                self.node.env.logger.info(f"{self.node.name}'s sifted key (Decimal): {transmitter_key}")
                self.sifted_keys.append(transmitter_key)

                # Check if it reaches the terminal condition
                if len(self.sifted_keys) >= self.key_num:
                    self.finish()  # transmitter finishes BB84 protocol
                    self.statistics_estimation()  # estimate sifted key rate and error rate

    def receiver_receive_classical_msg(self, msg: "BB84.Message", **kwargs) -> None:
        r"""Receiver receives a BB84 message from the transmitter.

        Args:
            msg (BB84.Message): message for BB84 protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        # Filter the message that is not intended for this instance
        if msg.data['type'].name not in ['START', 'BASES']:
            return

        if msg.data['type'] == BB84.Message.Type.START:
            self.node.env.logger.debug(f"{self.node.name} received message 'START' from {self.peer.name}")

            clock = msg.data['clock']  # time for photon reception
            frequency = msg.data['source_frequency']  # frequency for beam splitter
            self.light_duration = msg.data['light_duration']
            self.emission_rest = msg.data['emission_rest']

            # Generate receiver's random measurement bases
            self.pulse_num = round(msg.data['light_duration'] * frequency * 1e-12)
            bases = list(np.random.choice([0, 1], self.pulse_num, p=self.rx_bases_ratio))
            self.rx_bases_list.append(bases)

            bases = [self.basis_encoding[basis] for basis in bases]  # map the bits to bases for measurement
            # Set parameters for polarization detector
            polar_detector = self.node.polar_detector[self.peer] if isinstance(self.node.polar_detector, dict)\
                else self.node.polar_detector

            polar_detector.set_beamsplitter(clock=clock, frequency=frequency, bases=bases)
            polar_detector.detectors[0].turn_on()
            polar_detector.detectors[1].turn_on()

            # Stop receiving photons after a light duration
            handler = EventHandler(self, "receiver_stop_reception")
            self.scheduler.schedule_at(clock + self.light_duration, handler)

        elif msg.data['type'] == BB84.Message.Type.BASES:
            photon_round = msg.data['round']

            # Check photon round to filter the outdated messages resulting from BB84 protocol's reusing
            if self.round < photon_round:
                self.node.env.logger.debug(f"Message with round {photon_round} was discarded"
                                           f" at local round {self.round}.")
                return

            self.node.env.logger.debug(f"{self.node.name} received bases from {self.peer.name} "
                                       f"(Round: {photon_round})")

            peer_bases = msg.data['peer_bases']
            bases = self.rx_bases_list.pop(0)  # pop the earliest bases for basis comparison
            bits = self.rx_bits_list.pop(0)
            matching_list = []
            for i, basis in enumerate(bases):
                # Record valid bits of matching indices
                if basis == peer_bases[i] and bits[i] is not None:
                    matching_list.append(i)
                    self.rx_key_bits.append(bits[i])

            # Tell the transmitter the matching indices
            matching_msg = BB84.Message(src=self.node,
                                        dst=self.peer,
                                        protocol=BB84,
                                        data={'type': BB84.Message.Type.MATCHING,
                                              'matching_list': matching_list,
                                              'round': photon_round})
            self.node.send_classical_msg(dst=self.peer, msg=matching_msg)

            # Check if the length of key bits is enough for a new key
            if len(self.rx_key_bits) >= self.key_length:
                self.node.env.logger.info(f"{self.node.name} generated a valid key with {self.peer.name}"
                                          f" (Round {photon_round})")
                self.node.env.logger.info(f"{self.node.name}'s sifted key: {self.rx_key_bits[0: self.key_length]}")

                # Store the generated key as a decimal integer
                receiver_key = int("".join(str(self.rx_key_bits.pop(0)) for _ in range(self.key_length)), base=2)
                self.node.env.logger.info(f"{self.node.name}'s sifted key (Decimal): {receiver_key}")
                self.sifted_keys.append(receiver_key)

                # Check if enough keys are generated
                if len(self.sifted_keys) >= self.key_num:
                    self.finish()  # receiver finishes BB84 protocol


class DecoyBB84(PrepareAndMeasure):
    r"""Class for the decoy-state BB84 key generation protocol.

    Attributes:
        basis_encoding (dict): map an int to a measurement basis
        intensities (dict): intensities of the photon source (signal, decoy, vacuum)

    Examples:
        The ``intensities`` attribute of DecoyBB84 protocol is a dictionary which stores the probabilities and
        mean photon numbers of three different state types (signal, decoy, vacuum).
        Here is an example of setting ``intensities`` attribute of the DecoyBB84 protocol.

        >>> decoy_bb84 = DecoyBB84("Decoy BB84")
        >>> decoy_bb84.set_intensity({"prob": [0.5, 0.25, 0.25], "mean_photon_num": [0.8, 0.1, 0]})

    Warning:
        This is an example of decoy-state protocol.
        We assume the classical and quantum channels have the same distance so that the propogation delays of
        classical and quantum messages are the same.
    """

    def __init__(self, name: str, peer=None):
        r"""Constructor for DecoyBB84 class.

        Args:
            name (str): name of the decoy-state BB84 protocol
            peer (QKDNode): peer node to implement the same task
        """
        super().__init__(name, peer)
        self.basis_encoding = {0: Basis.Z(), 1: Basis.X()}
        self.intensities = {}
        self.tx_state_pos_list = []

    @unique
    class StateType(Enum):
        r"""Class for the state types of emitted pulses.
        """

        SIGNAL = 0
        DECOY = 1
        VACUUM = 2

    class Message(ClassicalMessage):
        r"""Class for the classical control messages in DecoyBB84 protocol.
        """

        def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
            r"""Constructor for Message class.

            Args:
                src (Node): source of Message
                dst (Node): destination of Message
                protocol (type): protocol of Message
                data (Dict): message content
            """
            super().__init__(src, dst, protocol, data)

        @unique
        class Type(Enum):
            r"""Class for Message types.
            """

            START = "Start"
            FINISH_EMISSION = "Finish emission"
            BASES = "Bases"
            MATCHING = "Matching"

    def set_intensity(self, intensities: dict) -> None:
        r"""Set the intensities of signal, decoy and vacuum pulses.

        Args:
            intensities (dict): intensity of the photon source (signal, decoy, vacuum)
        """
        self.intensities = intensities

    def start(self, **kwargs) -> None:
        r"""Start the decoy-state BB84 protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        assert self.node is not None and self.peer is not None, f"Should load the protocol stack to a node first!"

        # Initial settings
        self.role = kwargs['role']
        if "key_num" in kwargs.keys():
            self.key_num = kwargs['key_num']
        if "key_length" in kwargs.keys():
            self.key_length = kwargs['key_length']

        self.status = KeyGeneration.Status.WORKING
        self.sifted_keys = []
        self.round = 0

        if self.role == PrepareAndMeasure.Role.TRANSMITTER:
            self.start_time = self.node.env.now  # record the start time of DecoyBB84 protocol
            if self.tx_bases_ratio is None:
                self.tx_bases_ratio = [0.5, 0.5]
            self.tx_bases_list = []
            self.tx_bits_list = []
            self.tx_key_bits = []
            self.transmitter_start_protocol()

        elif self.role == PrepareAndMeasure.Role.RECEIVER:
            if self.rx_bases_ratio is None:
                self.rx_bases_ratio = [0.5, 0.5]
            self.rx_bases_list = []
            self.rx_bits_list = []
            self.rx_key_bits = []

    def transmitter_start_protocol(self) -> None:
        r"""Transmitter starts the decoy-state BB84 protocol.

        Transmitter sends a classical message to the receiver for device synchronization
        and prepares to emit photon pulses.
        """
        self.node.env.logger.info(f"{self.node.name} started decoy-state BB84 protocol with {self.peer.name}")

        self.pulse_num = round(self.key_length / self.node.photon_source.mean_photon_num)
        self.light_duration = round(self.pulse_num / self.node.photon_source.frequency * 1e12)
        self.emission_rest = round(1e12 / self.node.photon_source.frequency)

        transmitter_emit_time = self.node.env.now + self.node.cchannel(self.peer).delay
        clock = transmitter_emit_time + self.node.qchannel(self.peer).delay  # clock time for photon reception

        start_msg = DecoyBB84.Message(src=self.node,
                                      dst=self.peer,
                                      protocol=DecoyBB84,
                                      data={'type': DecoyBB84.Message.Type.START,
                                            'clock': clock,
                                            'source_frequency': self.node.photon_source.frequency,
                                            'light_duration': self.light_duration,
                                            'emission_rest': self.emission_rest})
        self.node.send_classical_msg(dst=self.peer, msg=start_msg)

        # Transmitter begins photon emission as soon as the receiver receives the message
        handler = EventHandler(self, "transmitter_start_emission")
        self.scheduler.schedule_after(self.node.cchannel(self.peer).delay, handler)

    def transmitter_start_emission(self) -> None:
        r"""Transmitter starts to emit photon pulses.

        Transmitter randomly generates bits and encoding bases for photon emission.
        """
        if self.status == KeyGeneration.Status.WORKING:
            self.round += 1
            self.node.env.logger.debug(f"{self.node.name} began photon emission for decoy-state BB84 with"
                                       f" {self.peer.name} (Round {self.round})")

            # Generate transmitter's random bases and bits
            bases = list(np.random.choice([0, 1], self.pulse_num, p=self.tx_bases_ratio))
            bits = list(np.random.choice([0, 1], self.pulse_num))  # uniform distribution
            self.tx_bases_list.append(bases)
            self.tx_bits_list.append(bits)

            bases = [self.basis_encoding[basis] for basis in bases]
            states = []
            for i, basis in enumerate(bases):
                # Prepare states |0>, |1>, |+>, |-> (in the form of density matrices) for photon emission
                states.append(basis[bits[i]] @ np.conjugate(basis[bits[i]]).T)

            # Randomly choose state types (signal, decoy or vacuum) according to the probabilities of intensities
            state_types = list(np.random.choice([self.StateType.SIGNAL, self.StateType.DECOY, self.StateType.VACUUM],
                                                self.pulse_num, p=self.intensities["prob"]))
            # Set related mean photon numbers
            mean_photon_num_list = [self.intensities["mean_photon_num"][state_type.value] for state_type in state_types]

            # Record the positions of different state types (signal, decoy and vacuum)
            state_pos_dict = {"signal": [], "decoy": [], "vacuum": []}
            for i, state_type in enumerate(state_types):
                if state_type == self.StateType.SIGNAL:
                    state_pos_dict["signal"].append(i)
                elif state_type == self.StateType.DECOY:
                    state_pos_dict["decoy"].append(i)
                elif state_type == self.StateType.VACUUM:
                    state_pos_dict["vacuum"].append(i)
            self.tx_state_pos_list.append(state_pos_dict)

            handler = EventHandler(self.node.photon_source, "emit", [self.peer, states, mean_photon_num_list])
            self.scheduler.schedule_now(handler)
            finish_emission_msg = DecoyBB84.Message(src=self.node,
                                                    dst=self.peer,
                                                    protocol=DecoyBB84,
                                                    data={'type': DecoyBB84.Message.Type.FINISH_EMISSION,
                                                          'round': self.round})
            handler = EventHandler(self.node, "send_classical_msg", [self.peer, finish_emission_msg])
            self.scheduler.schedule_after(self.light_duration, handler)

            # Schedule next round of photon emission
            handler = EventHandler(self, "transmitter_start_emission")
            # Rest for a while to avoid scheduling the event during cool-down period of the photon source
            self.scheduler.schedule_after(self.light_duration + self.emission_rest, handler)

    def receiver_stop_reception(self) -> None:
        r"""Receiver stops photon reception.

        Receiver pops the measurement outcomes and resets the measurement bases to
        prepare for receiving photon pulses in the next round.
        """
        if self.status == KeyGeneration.Status.WORKING:
            self.round += 1
            self.node.env.logger.debug(f"{self.node.name} finished receiving photons from {self.peer.name}"
                                       f" (Round {self.round})")

            bases = self.rx_bases_list.pop(0)
            bits = self.receiver_get_bits()
            self.rx_bits_list.append(bits)
            for i, bit in enumerate(bits):
                if bit is None:  # if the bit is invalid, set the related basis as invalid
                    bases[i] = None

            # Send the measurement bases to the transmitter
            bases_msg = DecoyBB84.Message(src=self.node,
                                          dst=self.peer,
                                          protocol=DecoyBB84,
                                          data={'type': DecoyBB84.Message.Type.BASES,
                                                'peer_bases': bases,
                                                'round': self.round})
            self.node.send_classical_msg(dst=self.peer, msg=bases_msg)

            # Reset measurement bases for the next round of photon reception
            new_bases = list(np.random.choice([0, 1], self.pulse_num, p=self.rx_bases_ratio))
            self.rx_bases_list.append(new_bases)
            clock = self.node.env.now + self.emission_rest  # set the next clock time

            new_bases = [self.basis_encoding[basis] for basis in new_bases]
            # Reset parameters for polarization detector
            self.node.polar_detector.set_beamsplitter(clock=clock, bases=new_bases)

        elif self.status == KeyGeneration.Status.DONE:
            self.node.polar_detector.detectors[0].turn_off()
            self.node.polar_detector.detectors[1].turn_off()

    def transmitter_receive_classical_msg(self, msg: "DecoyBB84.Message", **kwargs) -> None:
        r"""Transmitter receives a DecoyBB84 message from the receiver.

        Args:
            msg (DecoyBB84.Message): message for DecoyBB84 protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        photon_round = msg.data['round']

        # if self.node.env.now < self.end_time:
        if msg.data['type'] == DecoyBB84.Message.Type.BASES:
            self.node.env.logger.debug(f"{self.node.name} received bases from {self.peer.name}"
                                       f" (Round {photon_round})")

            peer_bases = msg.data['peer_bases']
            # Pop the earliest bases and bits for index comparison
            bases = self.tx_bases_list.pop(0)
            bits = self.tx_bits_list.pop(0)
            state_pos = self.tx_state_pos_list.pop(0)

            matching_list = []  # store indices of matching bases
            # Sift out mismatching bases
            for ix, basis in enumerate(bases):
                if peer_bases[ix] is not None and basis == peer_bases[ix]:  # store the bases indices if matching
                    matching_list.append(ix)
                else:  # remove mismatching elements
                    if ix in state_pos["signal"]:
                        state_pos["signal"].remove(ix)  # remove mismatching signal states
                    elif ix in state_pos["decoy"]:
                        state_pos["decoy"].remove(ix)  # remove mismatching decoy states
                    elif ix in state_pos["vacuum"]:
                        state_pos["vacuum"].remove(ix)  # remove mismatching vacuum states

            # Randomly choose 10 percent of signal states for testing
            testing_signal_pos = list(np.random.choice(state_pos["signal"], round(0.1 * len(state_pos["signal"])),
                                                       replace=False))
            decoy_bits = []
            vacuum_bits = []
            testing_signal_bits = []
            # Sift key bits and bits for testing (decoy bits, vacuum bits, 10% signal bits)
            for ix, bit in enumerate(bits):
                if ix in state_pos["signal"]:
                    if ix in testing_signal_pos:
                        testing_signal_bits.append(bit)
                    else:
                        self.tx_key_bits.append(bit)
                elif ix in state_pos["decoy"]:
                    decoy_bits.append(bit)
                elif ix in state_pos["vacuum"]:
                    vacuum_bits.append(bit)

            # Send matching bases and testing bits with their positions to the receiver
            matching_msg = DecoyBB84.Message(src=self.node,
                                             dst=self.peer,
                                             protocol=DecoyBB84,
                                             data={'type': DecoyBB84.Message.Type.MATCHING,
                                                   'matching_list': matching_list,
                                                   'decoy_pos': state_pos["decoy"],
                                                   'vacuum_pos': state_pos["vacuum"],
                                                   "testing_signal_pos": testing_signal_pos,
                                                   "decoy_bits": decoy_bits,
                                                   "vacuum_bits": vacuum_bits,
                                                   "testing_signal_bits": testing_signal_bits,
                                                   'round': photon_round})
            self.node.send_classical_msg(dst=self.peer, msg=matching_msg)

            # Check if the length of key bits is enough for a new key
            if len(self.tx_key_bits) >= self.key_length:
                self.node.env.logger.info(f"{self.node.name} generated a valid key with {self.peer.name}"
                                          f" (Round {photon_round})")
                self.node.env.logger.info(f"{self.node.name}'s sifted key: {self.tx_key_bits[0: self.key_length]}")

                # Store the generated key as a decimal integer
                transmitter_key = int("".join(str(self.tx_key_bits.pop(0)) for _ in range(self.key_length)), base=2)
                self.node.env.logger.info(f"{self.node.name}'s sifted key (Decimal): {transmitter_key}")
                self.sifted_keys.append(transmitter_key)

                # Check if it reaches the terminal condition
                if len(self.sifted_keys) >= self.key_num:
                    self.finish()  # transmitter finishes DecoyBB84 protocol
                    # Calculate the overall sifted key rate in kilobit per second
                    self.key_rate_estimation()

    def receiver_receive_classical_msg(self, msg: "DecoyBB84.Message", **kwargs) -> None:
        r"""Receiver receives a DecoyBB84 message from the transmitter.

        Args:
            msg (DecoyBB84.Message): message for DecoyBB84 protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        if msg.data['type'] == DecoyBB84.Message.Type.START:
            self.node.env.logger.debug(f"{self.node.name} received message 'START' from {self.peer.name}")

            clock = msg.data['clock']  # time for photon reception
            frequency = msg.data['source_frequency']  # frequency for beam splitter
            self.light_duration = msg.data['light_duration']
            self.emission_rest = msg.data['emission_rest']

            # Generate receiver's random measurement bases
            self.pulse_num = round(msg.data['light_duration'] * frequency * 1e-12)
            bases = list(np.random.choice([0, 1], self.pulse_num, p=self.rx_bases_ratio))
            self.rx_bases_list.append(bases)

            bases = [self.basis_encoding[basis] for basis in bases]  # map the bits to bases for measurement
            # Set parameters for polarization detector
            self.node.polar_detector.set_beamsplitter(clock=clock, frequency=frequency, bases=bases)
            self.node.polar_detector.detectors[0].turn_on()
            self.node.polar_detector.detectors[1].turn_on()

        elif msg.data['type'] == DecoyBB84.Message.Type.FINISH_EMISSION:
            # Stop receiving photons after a light duration
            handler = EventHandler(self, "receiver_stop_reception")
            self.scheduler.schedule_now(handler)

        elif msg.data['type'] == DecoyBB84.Message.Type.MATCHING:
            photon_round = msg.data['round']
            self.node.env.logger.debug(f"{self.node.name} received matching indices from {self.peer.name} "
                                       f"(Round: {photon_round})")

            matching_list = msg.data['matching_list']
            decoy_pos = msg.data['decoy_pos']
            vacuum_pos = msg.data['vacuum_pos']
            testing_signal_pos = msg.data['testing_signal_pos']
            transmitter_decoy_bits = msg.data['decoy_bits']
            transmitter_vacuum_bits = msg.data['vacuum_bits']
            transmitter_testing_signal_bits = msg.data['testing_signal_bits']

            bits = self.rx_bits_list.pop(0)
            decoy_bits = []
            vacuum_bits = []
            testing_signal_bits = []
            # Extract bits of different state types for testing
            for i in matching_list:
                if i in decoy_pos:
                    decoy_bits.append(i)
                elif i in vacuum_pos:
                    vacuum_bits.append(i)
                elif i in testing_signal_pos:
                    testing_signal_bits.append(i)

            # Parameter estimation is done by simply comparing the decoy, vacuum and testing signal bits here.
            num_testing_errors = 0
            testing_error_rate = 0
            # Parameter estimation: calculate errors for eavesdropping detection
            for decoy_bit, transmitter_decoy_bit in zip(decoy_bits, transmitter_decoy_bits):
                if decoy_bit ^ transmitter_decoy_bit == 1:
                    num_testing_errors += 1
            for vacuum_bit, transmitter_vacuum_bit in zip(vacuum_bits, transmitter_vacuum_bits):
                if vacuum_bit ^ transmitter_vacuum_bit == 1:
                    num_testing_errors += 1
            for testing_signal_bit, transmitter_testing_signal_bit in \
                    zip(testing_signal_bits, transmitter_testing_signal_bits):
                if testing_signal_bit ^ transmitter_testing_signal_bit == 1:
                    num_testing_errors += 1
            # Avoid ZeroDivisionError
            if len(decoy_bits) + len(vacuum_bits) + len(testing_signal_bits) > 0:
                testing_error_rate = num_testing_errors / \
                                     (len(decoy_bits) + len(vacuum_bits) + len(testing_signal_bits))
            self.node.env.logger.debug(f"testing error rate: {testing_error_rate} (Round {photon_round})")

            # Set the remained signal bits as key bits
            key_bits_pos = list(set(matching_list) - set(decoy_pos) - set(vacuum_pos) - set(testing_signal_pos))
            key_bits_pos.sort()
            self.rx_key_bits.extend([bits[i] for i in key_bits_pos])

            # Check if the length of key bits is enough for a new key
            if len(self.rx_key_bits) >= self.key_length:
                self.node.env.logger.info(f"{self.node.name} generated a valid key with {self.peer.name}"
                                          f" (Round {photon_round})")
                self.node.env.logger.info(f"{self.node.name}'s sifted key: {self.rx_key_bits[0: self.key_length]}")

                # Store the generated key as a decimal integer
                receiver_key = int("".join(str(self.rx_key_bits.pop(0)) for _ in range(self.key_length)), base=2)
                self.node.env.logger.info(f"{self.node.name}'s sifted key (Decimal): {receiver_key}")
                self.sifted_keys.append(receiver_key)

                # Check if enough keys are generated
                if len(self.sifted_keys) >= self.key_num:
                    self.finish()  # receiver finishes BB84 protocol
                    self.error_rate_estimation()  # estimate the error of sifted keys


class B92(PrepareAndMeasure):
    r"""Class for the B92 key generation protocol.
    """

    def __init__(self, name: str, peer=None):
        r"""Constructor for B92 class.

        Args:
            name (str): name of the B92 protocol
            peer (QKDNode): peer node to implement the same task
        """
        super().__init__(name, peer)

    def transmitter_receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        pass

    def receiver_receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        pass


class SixState(PrepareAndMeasure):
    r"""Class for the six-state key generation protocol.
    """

    def __init__(self, name: str, peer=None):
        r"""Constructor for SixState class.

        Args:
            name (str): name of the six-state protocol
            peer (QKDNode): peer node to implement the same task
        """
        super().__init__(name, peer)

    def transmitter_receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        pass

    def receiver_receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        pass



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for quantum noises.
"""

from typing import List
import numpy
from qcompute_qnet.quantum.gate import Gate

__all__ = [
    "Noise"
]


class Noise:
    r"""Class to obtain quantum noise models.
    """

    @classmethod
    def BitFlip(cls, prob: float) -> List[numpy.ndarray]:
        r"""Kraus operators of a quantum bit flip noise.

        Such a noise has Kraus operators:

        .. math::

            E_0 = \sqrt{1-p} I, \quad
            E_1 = \sqrt{p} X.

        Args:
            prob (float): probability of a bit flip

        Returns:
            List[numpy.ndarray]: a list of kraus operators for the noise
        """
        assert 0 <= prob <= 1, "`prob` should be between 0 and 1."
        return [numpy.sqrt(1 - prob) * Gate.I(), numpy.sqrt(prob) * Gate.X()]

    @classmethod
    def PhaseFlip(cls, prob: float) -> List[numpy.ndarray]:
        r"""Kraus operators of a quantum phase flip noise.

        Such a noise has Kraus operators:

        .. math::

            E_0 = \sqrt{1-p} I, \quad
            E_1 = \sqrt{p} Z.

        Args:
            prob (float): probability of a phase flip

        Returns:
            List[numpy.ndarray]: a list of kraus operators for the noise
        """
        assert 0 <= prob <= 1, "`prob` should be between 0 and 1."
        return [numpy.sqrt(1 - prob) * Gate.I(), numpy.sqrt(prob) * Gate.Z()]

    @classmethod
    def BitPhaseFlip(cls, prob: float) -> List[numpy.ndarray]:
        r"""Kraus operators of a quantum bit-phase flip noise.

        Such a noise has Kraus operators:

        .. math::

            E_0 = \sqrt{1-p} I, \quad
            E_1 = \sqrt{p} Y.

        Args:
            prob (float): probability of a bit-phase flip

        Returns:
            List[numpy.ndarray]: a list of kraus operators for the noise
        """
        assert 0 <= prob <= 1, "`prob` should be between 0 and 1."
        return [numpy.sqrt(1 - prob) * Gate.I(), numpy.sqrt(prob) * Gate.Y()]

    @classmethod
    def AmplitudeDamping(cls, prob: float) -> List[numpy.ndarray]:
        r"""Kraus operators of a quantum amplitude damping noise.

        Such a noise has Kraus operators:

        .. math::

            E_0 =
            \begin{bmatrix}
                1 & 0 \\
                0 & \sqrt{1-p}
            \end{bmatrix}, \quad
            E_1 =
            \begin{bmatrix}
                0 & \sqrt{p} \\
                0 & 0
            \end{bmatrix}.

        Args:
            prob (float): damping probability

        Returns:
            List[numpy.ndarray]: a list of kraus operators for the noise
        """
        assert 0 <= prob <= 1, "`prob` should be between 0 and 1."
        return [numpy.array([[1, 0], [0, numpy.sqrt(1 - prob)]], dtype=complex),
                numpy.array([[0, numpy.sqrt(prob)], [0, 0]], dtype=complex)]

    @classmethod
    def PhaseDamping(cls, prob: float) -> List[numpy.ndarray]:
        r"""Kraus operators of a quantum phase damping noise.

        Such a noise has Kraus operators:

        .. math::

            E_0 =
            \begin{bmatrix}
                1 & 0 \\
                0 & \sqrt{1-p}
            \end{bmatrix}, \quad
            E_1 =
            \begin{bmatrix}
                0 & 0 \\
                0 & \sqrt{p}
            \end{bmatrix}.

        Args:
            prob (float): probability of a bit flip

        Returns:
            List[numpy.ndarray]: a list of kraus operators for the noise
        """
        assert 0 <= prob <= 1, "`prob` should be between 0 and 1."
        return [numpy.array([[1, 0], [0, numpy.sqrt(1 - prob)]], dtype=complex),
                numpy.array([[0, 0], [0, numpy.sqrt(prob)]], dtype=complex)]

    @classmethod
    def Depolarizing(cls, prob: float) -> List[numpy.ndarray]:
        r"""Kraus operators of a quantum depolarizing noise.

        Such a noise has Kraus operators:

        .. math::

            E_0 = \sqrt{1-p} I, \quad
            E_1 = \sqrt{p/3} X, \quad
            E_2 = \sqrt{p/3} Y, \quad
            E_3 = \sqrt{p/3} Z.

        Args:
            prob (float): parameter of the depolarizing noise

        Returns:
            List[numpy.ndarray]: a list of kraus operators for the noise
        """
        assert 0 <= prob <= 1, "`prob` should be between 0 and 1."
        return [numpy.sqrt(1 - prob) * Gate.I(),
                numpy.sqrt(prob / 3) * Gate.X(),
                numpy.sqrt(prob / 3) * Gate.Y(),
                numpy.sqrt(prob / 3) * Gate.Z()]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for nodes in a network.
"""

from abc import ABC
from typing import Any
from qcompute_qnet.core.des import Entity
from qcompute_qnet.functionalities.mobility import Mobility
from qcompute_qnet.protocols.protocol import ProtocolStack
from qcompute_qnet.messages.message import ClassicalMessage, QuantumMessage

__all__ = [
    "Node",
    "Satellite"
]


class Node(Entity):
    r"""Class for creating a node.

    Attributes:
        links (Dict[Node, Link]): links connecting the node with other nodes
        protocol_stack (ProtocolStack): protocol stack installed
        functionalities (list): functionalities of the node
        is_mobile (bool): whether the node is a mobile node
    """

    def __init__(self, name: str, env=None):
        r"""Constructor for Node class.

        Args:
            name (str): name of the node
            env (DESEnv): related discrete-event simulation environment
        """
        super().__init__(name, env)
        self.links = {}
        self.protocol_stack = None
        self.functionalities = []
        self.is_mobile = False

    def init(self) -> None:
        r"""Node initialization.

        This method will do a sanity check to confirm that the node is installed to a ``Network``.
        """
        assert self.owner != self, f"The node {self.name} should be installed to a 'Network' first!"

    def load_protocol(self, protocol_stack: "ProtocolStack") -> None:
        r"""Load a protocol stack to the node.

        ``load_protocol`` method will set the node for all protocols in the protocol stack and synchronize
        their scheduler's simulation environment as the one of node.

        Args:
            protocol_stack (ProtocolStack): protocol stack to load
        """
        self.protocol_stack = protocol_stack
        protocol_stack.owner = self
        protocol_stack.sync_env(self)

    def assign_functionality(self, functionality: Any) -> None:
        r"""Assign a functionality to the node.

        Args:
            functionality (Any): functionality to assign
        """
        self.functionalities.append(functionality)

        from qcompute_qnet.functionalities.mobility import Mobility
        if isinstance(functionality, Mobility):
            self.is_mobile = True

    def cchannel(self, dst: "Node") -> "ClassicalChannel":
        r"""Find the classical channel with a given destination.

        Args:
            dst (Node): destination node of the classical channel

        Returns:
            ClassicalChannel: the classical channel with the given destination
        """
        return self.links[dst].cchannel(dst)

    def qchannel(self, dst: "Node") -> "QuantumChannel":
        r"""Find the quantum channel with a given destination.

        Args:
            dst (Node): destination node of the quantum channel

        Returns:
            QuantumChannel: the quantum channel with the given destination
       """
        return self.links[dst].qchannel(dst)

    def send_classical_msg(self, dst: "Node", msg: "ClassicalMessage", priority=None) -> None:
        r"""Send a classical message through the classical channel.

        Args:
            dst (Node): destination of the message
            msg (ClassicalMessage): classical message to send
            priority (int): priority of the transmission event
        """
        self.cchannel(dst).transmit(msg, priority)

    def send_quantum_msg(self, dst: "Node", msg: "QuantumMessage", priority=None) -> None:
        r"""Send a quantum message through the quantum channel.

        Args:
            dst (Node): destination of the message
            msg (QuantumMessage): quantum message to send
            priority (int): priority of the transmission event
        """
        self.qchannel(dst).transmit(msg, priority)

    def receive_classical_msg(self, src: "Node", msg: "ClassicalMessage") -> None:
        r"""Receive a classical message from the classical channel.

        The source node is inferred by the receiver and is not part of the transmitted message.

        Args:
            src (Node): source of the classical message
            msg (ClassicalMessage): classical message to receive
        """
        pass

    def receive_quantum_msg(self, src: "Node", msg: "QuantumMessage") -> None:
        r"""Receive a quantum message from the quantum channel.

        The source node is inferred by the receiver and is not part of the transmitted message.

        Args:
            src (Node): source of the quantum message
            msg (QuantumMessage): quantum message to receive
        """
        pass

    def start(self, **kwargs) -> None:
        r"""Start the protocol stack of the node.

        Args:
            **kwargs: keyword arguments to start the protocol stack
        """
        assert self.protocol_stack is not None, f"Should load a protocol stack first."
        self.protocol_stack.start(**kwargs)


class Satellite(Node, ABC):
    r"""Abstract class for satellite nodes.

    A satellite node is a mobile node that keeps moving on its orbit.
    """

    def __init__(self, name: str, env=None):
        r"""Constructor for Satellite class.

        Args:
            name (str): name of the satellite node
            env (DESEnv): related discrete-event simulation environment
        """
        super().__init__(name, env)
        self.mobility = Mobility()
        self.assign_functionality(self.mobility)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for messages in quantum key distribution.
"""

from enum import Enum, unique
from typing import Dict
from qcompute_qnet.messages.message import ClassicalMessage

__all__ = [
    "QKDMessage"
]


class QKDMessage(ClassicalMessage):
    r"""Class for the classical control messages in quantum key distribution.
    """

    def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
        r"""Constructor for QKDMessage class.

        Args:
            src (Node): source of QKDMessage
            dst (Node): destination of QKDMessage
            protocol (type): protocol of QKDMessage
            data (Dict): message content
        """
        super().__init__(src, dst, protocol, data)

    @unique
    class Type(Enum):
        r"""Class for QKDMessage types.
        """

        REQUEST = "Request"
        ACCEPT = "Accept"
        READY = "Ready"
        CIPHERTEXT = "Ciphertext"
        ACKNOWLEDGE = "Acknowledge"
        DONE = "Done"
        # Types especially for QKD with resource management
        PATH = "Path"
        RESERVE = "Reserve"
        POP = "Pop"
        CONFIRM = "Confirm"
        REJECT = "Reject"
        GENERATE = "Generate"



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for communication links in a network.
"""

from qcompute_qnet.core.des import Entity
from qcompute_qnet.devices.channel import ClassicalChannel, QuantumChannel, DuplexClassicalFiberChannel
from qcompute_qnet.devices.channel import DuplexQuantumFiberChannel, ClassicalFiberChannel, QuantumFiberChannel

__all__ = [
    "Link"
]


class Link(Entity):
    r"""Class for creating a link.

    Attributes:
        ends (Tuple[Node]): ends connected by the link

    Note:
        ``Link`` is an abstract concept indicating the connectivity between two nodes.
        The information transmission is realized by its channel components.
        So we should install classical and quantum channels to a link before running the simulation.

    Examples:
        We can create a link between two nodes (e.g. alice and bob) by two approaches.
        The first approach is to instantiate a ``Link`` with given ``name``.
        Then using the ``connect`` method to make a connection to alice and bob.

        >>> l_ab = Link(name="Link_Alice_Bob")
        >>> l_ab.connect(alice, bob)

        The second approach is to assign the two ends when instantiating.
        This will make your code more compact.

        >>> l_ab = Link(name="Link_Alice_Bob", ends=(alice, bob))

        Once we create a link, we can install some specific channels to it. Taking BB84 protocol as an example,
        we need a classical channel from alice to bob and a classical channel from bob to alice.
        We will also need a quantum channel from alice to bob. This can be done as follows:

        >>> l_ab = Link(name="Link_Alice_Bob", ends=(alice, bob))
        >>> c_a2b = ClassicalFiberChannel(name="c_a2b", distance=1e3, sender=alice, receiver=bob)
        >>> c_b2a = ClassicalFiberChannel(name="c_b2a", distance=1e3, sender=bob, receiver=alice)
        >>> q_a2b = QuantumFiberChannel(name="q_a2b", distance=1e3, sender=alice, receiver=bob)
        >>> l_ab.install([c_a2b, c_b2a, q_a2b])
    """

    def __init__(self, name: str, env=None, ends=None, distance=None):
        r"""Constructor for Link class.

        Args:
            name (str): name of the link
            env (DESEnv, optional): discrete-event simulation environment
            ends (Tuple[Node], optional): ends connected by the link
            distance (int, optional): length of the link in meters
        """

        super().__init__(name, env)
        self.ends = (None, None)
        if distance is not None:
            assert ends is not None, f"Should set the ends of the link once specify its distance."
            cchannel1_2 = ClassicalChannel(
                f"cchannel_{ends[0]}2{ends[1]}", distance=distance, sender=ends[0], receiver=ends[1]
            )
            cchannel2_1 = ClassicalChannel(
                f"cchannel_{ends[1]}2{ends[0]}", distance=distance, sender=ends[1], receiver=ends[0]
            )
            qchannel1_2 = QuantumChannel(
                f"qchannel_{ends[0]}2{ends[1]}", distance=distance, sender=ends[0], receiver=ends[1]
            )
            qchannel2_1 = QuantumChannel(
                f"qchannel_{ends[1]}2{ends[0]}", distance=distance, sender=ends[1], receiver=ends[0]
            )
            self.install([cchannel1_2, cchannel2_1, qchannel1_2, qchannel2_1])
        if ends is not None:
            self.connect(ends[0], ends[1])

    def init(self) -> None:
        r"""Link initialization.

        This method will do a sanity check to confirm that the link is installed to a ``Network``.
        """
        assert self.owner != self, f"The link {self.name} should be installed to a 'Network' first!"

    def cchannel(self, dst: "Node") -> "ClassicalChannel":
        r"""Find the classical channel with a given destination.

        Args:
            dst (Node): destination node of the classical channel

        Returns:
            ClassicalChannel: the classical channel with the given destination
        """
        for component in self.components:
            if isinstance(component, ClassicalChannel) and component.receiver == dst:
                return component
            elif isinstance(component, DuplexClassicalFiberChannel) and component.components[0].receiver == dst:
                return component.components[0]
            elif isinstance(component, DuplexClassicalFiberChannel) and component.components[1].receiver == dst:
                return component.components[1]
        raise Exception("No such a classical channel in the link!")

    def qchannel(self, dst: "Node") -> "QuantumChannel":
        r"""Find the quantum channel with a given destination.

        Args:
            dst (Node): destination node of the quantum channel

        Returns:
            QuantumChannel: the quantum channel with the given destination
        """
        for component in self.components:
            if isinstance(component, QuantumChannel) and component.receiver == dst:
                return component
            elif isinstance(component, DuplexQuantumFiberChannel) and component.components[0].receiver == dst:
                return component.components[0]
            elif isinstance(component, DuplexQuantumFiberChannel) and component.components[1].receiver == dst:
                return component.components[1]
        raise Exception("No such a quantum channel in the link!")

    def connect(self, node1: "Node", node2: "Node") -> None:
        r"""Connect the link to two given nodes.

        Args:
            node1 (Node): one end of the link
            node2 (Node): the other end of the link
        """
        assert self.ends[0] is None and self.ends[1] is None, \
            f"Link has been connected to '{self.ends[0].name}' and '{self.ends[1].name}'."

        self.ends = (node1, node2)
        node1.links[node2] = self
        node2.links[node1] = self



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for quantum measurement bases.
"""

from argparse import ArgumentTypeError
import numpy
from numpy import pi
from qcompute_qnet.quantum.gate import Gate
from qcompute_qnet.quantum.state import One, Zero, Plus, Minus

__all__ = [
    "Basis"
]


class Basis:
    r"""Class to obtain a measurement basis.
    """

    @classmethod
    def X(cls) -> numpy.ndarray:
        r"""Return X basis.

        Returns:
            numpy.ndarray: X basis
        """
        return numpy.array([Plus.SV, Minus.SV])

    @classmethod
    def Y(cls) -> numpy.ndarray:
        r"""Return Y basis.

        Returns:
            numpy.ndarray: Y basis
        """
        return numpy.array([Gate.Rz(pi / 2) @ Plus.SV, Gate.Rz(pi / 2) @ Minus.SV])
    
    @classmethod
    def Z(cls) -> numpy.ndarray:
        r"""Return Z basis.

        Returns:
            numpy.ndarray: Z basis
        """
        return numpy.array([Zero.SV, One.SV])

    @classmethod
    def Plane(cls, plane: str, theta: float) -> numpy.ndarray:
        r"""Single-qubit measurement basis on a given plane.

        .. math::

            M^{XY}(\theta) = \{R_{z}(\theta)|+\rangle, R_{z}(\theta)|-\rangle\}

            M^{YZ}(\theta) = \{R_{x}(\theta)|0\rangle, R_{x}(\theta)|1\rangle\}

            M^{XZ}(\theta) = \{R_{y}(\theta)|0\rangle, R_{y}(\theta) |1\rangle\}

        Args:
            plane (str): measurement plane, can be 'XY', 'YZ' or 'XZ'
            theta (float): measurement angle

        Returns:
            numpy.ndarray: measurement basis
        """
        if plane == 'XY':  # XY plane measurement basis
            return numpy.array([Gate.Rz(theta) @ Plus.SV, Gate.Rz(theta) @ Minus.SV])

        elif plane == 'YZ':  # YZ plane measurement basis
            return numpy.array([Gate.Rx(theta) @ Zero.SV, Gate.Rx(theta) @ One.SV])

        elif plane == 'XZ':  # XZ plane measurement basis
            return numpy.array([Gate.Ry(theta) @ Zero.SV, Gate.Ry(theta) @ One.SV])

        else:
            raise ArgumentTypeError(f"Input {plane} should be 'XY', 'YZ' or 'XZ'.")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for discrete-event simulation environment with extra functionalities.
"""

from enum import Enum
from typing import List, Union

from qcompute_qnet.core.des import DESEnv

__all__ = [
    "QuantumEnv"
]


class QuantumEnv(DESEnv):
    r"""Class for discrete-event simulation environment with extra functionalities.

    Important:
        A ``QuantumEnv`` is a ``DESEnv`` with extra functionalities that can implement quantum circuits
        with virtual or real quantum backends.

        These quantum circuits are automatically generated from quantum network protocols.
    """

    def __init__(self, name: str, default=False):
        r"""Constructor for QuantumEnv class.

        Args:
            name (str): name of the discrete-event simulation environment
            default (bool): whether to set the current environment as default
        """
        super().__init__(name, default)

    def run(self, shots=1, backend=None, token=None, print_cir=True,
            end_time=None, logging=False, summary=True) -> Union[dict, List[dict]]:
        r"""Run the simulation and the quantum circuit with a given backend.

        Args:
            shots (int): number of circuit samples for a single circuit
            backend (Enum): backend to run the quantum circuit
            token (str): your token for QCompute backend
            print_cir (bool): whether to print the circuit
            end_time (float): end time of the simulation in picoseconds
            logging (bool): whether to output the simulation log file
            summary (bool): whether to print the simulation report on the terminal

        Returns:
            Union[dict, List[dict]]: circuit results
        """
        super().run(end_time, logging, summary)

        if len(self.network.circuits) == 0:  # single circuit
            cir = self.network.default_circuit
            cir.defer_measurement()
            if print_cir:
                cir.print_circuit(color=True)

            results = cir.run(shots, backend, token)

        else:  # multiple circuits
            circuits = {self.network.circuits.pop(0): 1}  # record the same circuits with their repeated times
            results = []

            for comp_cir in self.network.circuits:
                diff = True
                for cir in circuits:
                    if comp_cir.is_equal(cir):
                        diff = False
                        circuits[cir] += 1
                        break
                if diff is True:
                    circuits[comp_cir] = 1
            self.network.circuits = list(circuits.keys())

            for cir in circuits.keys():
                cir.defer_measurement()
                if print_cir:
                    cir.print_circuit(color=True)

                result = cir.run(circuits[cir], backend, token)
                results.append(result)

        return results

    def reset(self) -> None:
        pass



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for measurement patterns used in the measurement-based quantum computation.
"""

from typing import Union, Any

__all__ = [
    "Pattern"
]


class Pattern:
    r"""Class for creating a measurement pattern.

    This class represents the measurement pattern in the MBQC model.
    Please refer to [The measurement calculus, arXiv: 0704.1263] for technical details.

    Attributes:
        name (str): pattern name
        space (list): space vertices
        input_ (list): input vertices
        output_ (list): output vertices
        commands (list): command list
    """

    def __init__(self, name: str, space: list, input_: list, output_: list, commands: list):
        r"""Constructor for Pattern class.

        Args:
            name (str): pattern name
            space (list): space vertices
            input_ (list): input vertices
            output_ (list): output vertices
            commands (list): command list
        """
        self.name = name
        self.space = space
        self.input_ = input_
        self.output_ = output_
        self.commands = commands

    class CommandE:
        r"""Class for creating an entanglement command.

        It entangles two adjacent vertices by operating a Control Z (CZ) gate on them.

        Attributes:
            which_qubits (list): a list of two vertices to be entangled
        """

        def __init__(self, which_qubits: list):
            r"""Constructor for CommandE class.

            Args:
                which_qubits (list): a pair of vertex labels
            """
            self.name = 'E'
            self.which_qubits = which_qubits

    class CommandM:
        r"""Class for creating a measurement command.

        It has five attributes including vertex label, primitive angle, measurement plane, 'domain_s', and 'domain_t'.
        Let :math:`\alpha` be the primitive measurement angle. After considering the dependencies to the
        measurement outcomes of some other vertices, the adaptive angle is calculated using the formula:

        .. math::

            \theta = (-1)^s \times \alpha + t \times \pi

        Note:
            'domain_s' and 'domain_t' are crucial concepts in the MBQC model.
            For detailed definitions, please refer to the reference [The measurement calculus, arXiv: 0704.1263].
            These two domains respectively record the effect of Pauli-X and Pauli-Z operators to the measurement angles.
            In other words, in order to calculate a measurement angle, the dependencies to the outcomes
            of some other vertices have to be taken into consideration in most cases.

        Warning:
            This command only supports measurement in the XY or YZ plane in the current version.

        Attributes:
            which_qubit (any): vertex label
            angle (Union[float, int]): primitive angle
            plane (str): measurement plane, can be 'XY' or 'YZ'
            domain_s (list): a list of vertices in 'domain_s' that have dependencies to this command
            domain_t (list): a list of vertices in 'domain_t' that have dependencies to this command
        """

        def __init__(self, which_qubit: Any, angle: Union[float, int], plane: str, domain_s: list, domain_t: list):
            r"""Constructor for CommandM class.

            Args:
                which_qubit (any): vertex label
                angle (Union[float, int]): primitive angle
                plane (str): measurement plane
                domain_s (list): a list of vertices in 'domain_s' that have dependencies to this command
                domain_t (list): a list of vertices in 'domain_t' that have dependencies to this command
            """
            assert isinstance(angle, float) or isinstance(angle, int), \
                f"Invalid measurement angle ({angle}) with the type: `{type(angle)}`!\n"\
                "Only `float` and `int` are supported as the type of measurement angle."

            self.name = 'M'
            self.which_qubit = which_qubit
            self.angle = angle
            self.plane = plane
            self.domain_s = domain_s
            self.domain_t = domain_t

    class CommandX:
        r"""Class for creating a Pauli-X correction command.

        Attributes:
            which_qubit (any): vertex label
            domain (list): a list of vertices that have dependencies to this command
        """

        def __init__(self, which_qubit: Any, domain: list):
            r"""Constructor for CommandX class.

            Args:
                which_qubit (any): vertex label
                domain (list): a list of vertices that have dependencies to this command
            """
            self.name = 'X'
            self.which_qubit = which_qubit
            self.domain = domain

    class CommandZ:
        r"""Class for creating a Pauli-Z correction command.

        Attributes:
            which_qubit (any): vertex label
            domain (list): a list of vertices that have dependencies to this command
        """

        def __init__(self, which_qubit: Any, domain: list):
            r"""Constructor for CommandZ class.

            Args:
                which_qubit (any): vertex label
                domain (list): a list of vertices that have dependencies to this command
            """
            self.name = 'Z'
            self.which_qubit = which_qubit
            self.domain = domain

    class CommandS:
        r"""Class for creating a signal shifting command.

        Note:
            Signal shifting is a unique operation in MBQC.
            It can be used to simplify the measurement commands by excluding the dependencies in 'domain_t'.

        Attributes:
            which_qubit (any): vertex label
            domain (list): a list of vertices that have dependencies to this command
        """

        def __init__(self, which_qubit: Any, domain: list):
            r"""Constructor for CommandS class.

            Args:
                which_qubit (any): vertex label
                domain (list): a list of vertices that have dependencies to this command
            """
            self.name = 'S'
            self.which_qubit = which_qubit
            self.domain = domain

    def print(self) -> None:
        r"""Print all commands in the command list.
        """
        print("-----------------------------------------------------------")
        print("                    Current Command List                   ")
        print("-----------------------------------------------------------")
        # Print commands list
        for cmd in self.commands:
            print('\033[91m' + 'Command:'.ljust(16) + cmd.name + '\033[0m')
            if cmd.name == 'E':
                print('which_qubits:'.ljust(15), cmd.which_qubits)
            elif cmd.name == 'M':
                print('which_qubit:'.ljust(15), cmd.which_qubit)
                print('plane:'.ljust(15), cmd.plane)
                print('angle:'.ljust(15), cmd.angle)
                print('domain_s:'.ljust(15), cmd.domain_s)
                print('domain_t:'.ljust(15), cmd.domain_t)
            else:
                print('which_qubit:'.ljust(15), cmd.which_qubit)
                print('domain:'.ljust(15), cmd.domain)
            print("-----------------------------------------------------------")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for different quantum circuit backends.
"""

from enum import Enum
import QCompute

__all__ = [
    "mbqc",
    "qcompute",
    "Backend"
]


class _QComputeBackend(Enum):
    r"""QCompute backends.
    """

    LocalBaiduSim2 = QCompute.BackendName.LocalBaiduSim2
    CloudBaiduSim2Water = QCompute.BackendName.CloudBaiduSim2Water
    CloudBaiduSim2Earth = QCompute.BackendName.CloudBaiduSim2Earth
    CloudBaiduSim2Thunder = QCompute.BackendName.CloudBaiduSim2Thunder
    CloudBaiduSim2Heaven = QCompute.BackendName.CloudBaiduSim2Heaven
    CloudBaiduSim2Wind = QCompute.BackendName.CloudBaiduSim2Wind
    CloudBaiduSim2Lake = QCompute.BackendName.CloudBaiduSim2Lake
    CloudAerAtBD = QCompute.BackendName.CloudAerAtBD
    CloudIoPCAS = QCompute.BackendName.CloudIoPCAS
    CloudBaiduQPUQian = QCompute.BackendName.CloudBaiduQPUQian


class _MBQCBackend(Enum):
    r"""MBQC backends.
    """

    StateVector = "StateVector"


class Backend:
    r"""Backends for quantum circuit implementation.
    """

    QCompute = _QComputeBackend
    MBQC = _MBQCBackend



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for log records.
"""

import logging

__all__ = [
    "LoggingFilter",
    "log_init",
]


class LoggingFilter(logging.Filter):
    r"""Class for the filter of log file.

    Attributes:
        env (DESEnv): discrete-event simulation environment to log
    """
    def __init__(self, env: "DESEnv"):
        r"""Constructor for LoggingFilter class.

        Args:
            env (DESEnv): discrete-event simulation environment to log
        """
        super().__init__()
        self.env = env

    def filter(self, record: logging.LogRecord) -> bool:
        r"""Filter the log records.

        Args:
            record (logging.LogRecord): an event to be logged

        Returns:
            bool: whether the specified record is to be logged
        """
        record.virtual_time = self.env.now

        return True


def log_init(env: "DESEnv", path=None, level="DEBUG") -> logging.Logger:
    r"""Initialization of a log file.

    Args:
        env (DESEnv): discrete-event simulation environment to log
        path (str): name and storing path of the log
        level (str): logging level of the logger

    Returns:
        logging.Logger: initialized logger
    """
    _log_path = f"./{env.name}.log" if path is None else path
    _logger = logging.getLogger(_log_path)
    _logger.setLevel(level)

    if env.logging:
        open(_log_path, "w").close()
        file_handler = logging.FileHandler(_log_path)
        log_format = "%(asctime)s:\t %(levelname)s\t %(virtual_time)s\t %(message)s"
        formatter = logging.Formatter(log_format)
        file_handler.setFormatter(formatter)
        file_filter = LoggingFilter(env)

        _logger.addHandler(file_handler)
        _logger.addFilter(file_filter)

    return _logger



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for photons.
"""

from qcompute_qnet.messages.message import QuantumMessage

__all__ = [
    "Photon"
]


class Photon(QuantumMessage):
    r"""Class for the simulation of photons.

    Note that ``Photon`` is a specific type of ``QuantumMessage``.

    Attributes:
        wavelength (float): wavelength of the photon in nanometers
        state (MixedState): quantum state of the photon
        index (int): index of the photon in a pulse sequence

    Note:
        The ``index`` is actually obtained from a time synchronization module, but we regard it as an attribute of
        the photon for simplicity.
    """

    def __init__(self, wavelength=None, state=None, index=None):
        r"""Constructor for Photon class.

        Args:
            wavelength (float): wavelength of the photon in nanometers
            state (MixedState): quantum state of the photon
            index (int): index of the photon in a pulse sequence
        """
        self.wavelength = wavelength
        self.state = state
        self.index = index



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for quantum registers.
"""

from typing import List, Union, Optional, Any
import numpy

from qcompute_qnet.core.des import Entity, EventHandler
from qcompute_qnet.quantum.circuit import Circuit
from qcompute_qnet.quantum.state import QuantumState

__all__ = [
    "QuantumRegister"
]


class QuantumRegister(Entity):
    r"""Quantum register is a logical device that stores and processes quantum information.

    Note:
        Since quantum states can be entangled :math:`\rho_{AB} \neq \rho_A \otimes \rho_B`,
        storing the local states :math:`\rho_A` and :math:`\rho_B` cannot faithfully represent all the information.
        So a quantum circuit is built up in the `Network` that reflects the evolution of the network.

    Attributes:
        size (int): size of the quantum register
        units (List[dict]): list for information about the qubits stored locally
        circuit_index (int): index of the current circuit related to the quantum register
    """

    def __init__(self, name: str, size: int, env: Optional["DESEnv"] = None):
        r"""Constructor for QuantumRegister class.

        Args:
            name (str): name of the quantum register
            size (int): size of the quantum register
            env (DESEnv, optional): discrete-event simulation environment
        """
        super().__init__(name, env)
        self.size = size

        self.units = []  # each unit stores a qubit, with its local address, quantum state and source
        for i in range(self.size):
            self.units.append({"address": i, "qubit": None, "identifier": self.owner, 'outcome': None})
        self.circuit_index = -1
        self.__op_counter = 0

    def init(self) -> None:
        r"""Initialization of the quantum register.
        """
        assert self.owner != self, f"{self.name} has no owner!"

    @property
    def __count(self) -> int:
        r"""Count the number of operations of the quantum register.

        Returns:
            int: index of the current operation in the quantum register
        """
        self.__op_counter += 1
        return self.__op_counter

    @property
    def circuit(self) -> "Circuit":
        r"""Get the quantum circuit of the network according to the circuit index.

        Important:
            If the index is specified before, this method will return the circuit with the index from the ``circuits``
            list of the network; otherwise it will return the default circuit of the network.

        Returns:
            Circuit: default circuit or circuit with the specified index
        """
        if self.circuit_index == -1:  # return the default circuit
            return self.env.network.default_circuit
        else:  # return the circuit with the specified index
            return self.env.network.circuits[self.circuit_index]

    def create_circuit(self, name: str) -> None:
        r"""Create a new circuit and add it to the ``circuits`` list of the network.

        Args:
            name (str): name of the new created circuit
        """
        self.env.network.circuits.append(Circuit(name=name))
        self.circuit_index += 1

    def init_unit(self, address: int) -> None:
        r"""Initialize the quantum register unit with the given address.

        Args:
            address (int): unit address of the quantum register
        """
        index = self.circuit.init_new_qreg_unit()
        self.units[address]['qubit'] = QuantumState(state=numpy.array([[index]]))

    def reset(self, address: int) -> None:
        r"""Reset the qubit with the given address.

        Args:
            address (int): local address of the qubit to reset
        """
        self.units[address]['qubit'] = None
        self.units[address]['identifier'] = self.owner

    def get_qubit(self, address: int) -> "QuantumState":
        r"""Get the qubit with the given address.

        Warnings:
            Due to the no-cloning theorem, a unit gets reset when we fetch its qubit.

        Args:
            address (int): address of the qubit

        Returns:
            QuantumState: quantum state of the qubit with the given address
        """
        qubit = self.units[address]['qubit']
        self.reset(address)

        return qubit

    def get_address(self, identifier: Any) -> int:
        r"""Get the qubit address with a given id.

        Args:
            identifier (Any): identity to fetch the qubit

        Returns:
            int: local address of the qubit that matches the id
        """
        for unit in self.units:
            if unit['identifier'] == identifier:
                return unit['address']

    def store_qubit(self, qubit: "QuantumState", identifier: Any) -> None:
        r"""Store a received qubit to the quantum register.

        Note:
            This method will store the qubit to a quantum register unit with the smallest available address.
            The identifier is used to uniquely identify the qubit. An identifier is required to fetch this qubit later.

        Args:
            qubit (QuantumState): received qubit
            identifier (Any): identity to store the qubit
        """
        assert any(unit['qubit'] is None for unit in self.units), f"No available units in {self.name}."

        for unit in self.units:
            if unit['qubit'] is None:
                unit['qubit'] = qubit
                unit['identifier'] = identifier
                return

    def __add_single_qubit_gate(self, operation: str, address: int, **params) -> None:
        r"""Add a single qubit gate to the circuit list.

        Args:
            operation (str): gate operation
            address (int): unit address of the quantum register
            **params (Any): gate parameters
        """
        if self.units[address]['qubit'] is None:
            self.init_unit(address)

        which_qubit = int(self.units[address]['qubit'].state[0][0])

        # Single-qubit gate with condition
        if "condition" in params.keys():
            # Parameterized single-qubit gate with condition
            if operation in ['rx', 'ry', 'rz']:
                handler = EventHandler(self.circuit, operation, [which_qubit, params['angle']],
                                       signature=self.owner, condition=params['condition'])
            elif operation == 'u3':
                handler = EventHandler(self.circuit, operation, [which_qubit, *params['angles']],
                                       signature=self.owner, condition=params['condition'])
            # Fixed single-qubit gate with condition
            else:
                handler = EventHandler(self.circuit, operation, [which_qubit], signature=self.owner, **params)
        # Single-qubit gate without condition
        else:
            # Parameterized single-qubit gate without condition
            if operation in ['rx', 'ry', 'rz']:
                handler = EventHandler(self.circuit, operation, [which_qubit, params['angle']], signature=self.owner)
            elif operation == 'u3':
                handler = EventHandler(self.circuit, operation, [which_qubit, *params['angles']], signature=self.owner)
            # Fixed single-qubit gate without condition (including measurement)
            else:
                handler = EventHandler(self.circuit, operation, [which_qubit], signature=self.owner)

        # Use the operation count as the priority so that the gate are executed in the correct order
        self.scheduler.schedule_now(handler, priority=self.__count)

    def __add_double_qubit_gate(self, operation: str, address: List[int], **params) -> None:
        r"""Add a double qubit gate to the circuit list.

        Args:
            operation (str): gate operation
            address (list): unit address of the quantum register in the order of [control, target]
            **params (Any): gate parameters
        """
        if self.units[address[0]]['qubit'] is None:
            self.init_unit(address[0])
        if self.units[address[1]]['qubit'] is None:
            self.init_unit(address[1])

        ctrl, targ = int(self.units[address[0]]['qubit'].state[0][0]), int(self.units[address[1]]['qubit'].state[0][0])

        # Parameterized double-qubit gate
        if operation in ['crx', 'cry', 'crz']:
            handler = EventHandler(self.circuit, operation, [[ctrl, targ], params['angle']], signature=self.owner)
        elif operation == 'cu3':
            handler = EventHandler(self.circuit, operation, [[ctrl, targ], *params['angles']], signature=self.owner)
        # Fixed double-qubit gate
        else:
            handler = EventHandler(self.circuit, operation, [[ctrl, targ]], signature=self.owner)

        self.scheduler.schedule_now(handler, priority=self.__count)

    def id(self, address: int) -> None:
        r"""Apply an identity gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
        """
        self.__add_single_qubit_gate('id', address)

    def h(self, address: int, **condition) -> None:
        r"""Apply a Hadamard gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            **condition (int): control qubit address
        """
        self.__add_single_qubit_gate('h', address, **condition)

    def x(self, address: int, **condition) -> None:
        r"""Apply a X gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            **condition (int): control qubit address
        """
        self.__add_single_qubit_gate('x', address, **condition)

    def y(self, address: int, **condition) -> None:
        r"""Apply a Y gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            **condition (int): control qubit address
        """
        self.__add_single_qubit_gate('y', address, **condition)

    def z(self, address: int, **condition) -> None:
        r"""Apply a Z gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            **condition (int): control qubit address
        """
        self.__add_single_qubit_gate('z', address, **condition)

    def s(self, address: int) -> None:
        r"""Apply a S gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
        """
        self.__add_single_qubit_gate('s', address)

    def t(self, address: int) -> None:
        r"""Apply a T gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
        """
        self.__add_single_qubit_gate('t', address)

    def rx(self, address: int, theta: Union[float, int], **condition) -> None:
        r"""Apply a rotation gate around x-axis on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            theta (Union[float, int]): rotation angle
            **condition (int): control qubit address
        """
        self.__add_single_qubit_gate('rx', address, angle=theta, **condition)

    def ry(self, address: int, theta: Union[float, int], **condition) -> None:
        r"""Apply a rotation gate around y-axis on the local qubit a the given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            theta (Union[float, int]): rotation angle
            **condition (int): control qubit address
        """
        self.__add_single_qubit_gate('ry', address, angle=theta, **condition)

    def rz(self, address: int, theta: Union[float, int], **condition) -> None:
        r"""Apply a rotation gate around z-axis on the local qubit with the given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            theta (Union[float, int]): rotation angle
            **condition (int): control qubit address
        """
        self.__add_single_qubit_gate('rz', address, angle=theta, **condition)

    def u3(self, address: int, theta: Union[float, int], phi: Union[float, int], gamma: Union[float, int],
           **condition) -> None:
        r"""Apply a rotation gate on the local qubit with a given address.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            theta (Union[float, int]): the rotation angle of the Rx gate
            phi (Union[float, int]): the rotation angle of the left Rz gate
            gamma (Union[float, int]): the rotation angle of the right Rz gate
            **condition (int): control qubit address
        """
        angles = [theta, phi, gamma]
        self.__add_single_qubit_gate('u3', address, angles=angles, **condition)

    def ch(self, address: List[int]) -> None:
        r"""Apply a Controlled-Hadamard gate on the local qubits with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
        """
        self.__add_double_qubit_gate('ch', address)

    def cx(self, address: List[int]) -> None:
        r"""Apply a Controlled-X gate on the local qubits with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
        """
        self.__add_double_qubit_gate('cx', address)

    def cnot(self, address: List[int]) -> None:
        r"""Apply a Controlled-NOT gate on the local qubits with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
        """
        self.__add_double_qubit_gate('cx', address)

    def cy(self, address: List[int]) -> None:
        r"""Apply a Controlled-Y gate on the local qubits with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
        """
        self.__add_double_qubit_gate('cy', address)

    def cz(self, address: List[int]) -> None:
        r"""Apply a Controlled-Z gate on the local qubits with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
        """
        self.__add_double_qubit_gate('cz', address)

    def crx(self, address: List[int], theta: Union[float, int]) -> None:
        r"""Apply a Controlled-rotation gate around x-axis on the local qubit with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
            theta (Union[float, int]): rotation angle
        """
        self.__add_double_qubit_gate('crx', address, angle=theta)

    def cry(self, address: List[int], theta: Union[float, int]) -> None:
        r"""Apply a Controlled-rotation gate around y-axis on the local qubit with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
            theta (Union[float, int]): rotation angle
        """
        self.__add_double_qubit_gate('cry', address, angle=theta)

    def crz(self, address: List[int], theta: Union[float, int]) -> None:
        r"""Apply a Controlled-rotation gate around z-axis on the local qubit with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
            theta (Union[float, int]): rotation angle
        """
        self.__add_double_qubit_gate('crz', address, angle=theta)

    def cu3(self, address: List[int],
            theta: Union[float, int], phi: Union[float, int], gamma: Union[float, int]) -> None:
        r"""Apply a Controlled-rotation gate on the local qubit with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
            theta (Union[float, int]): the rotation angle of the Rx gate
            phi (Union[float, int]): the rotation angle of the left Rz gate
            gamma (Union[float, int]): the rotation angle of the right Rz gate
        """
        angles = [theta, phi, gamma]
        self.__add_double_qubit_gate('cu3', address, angles=angles)

    def swap(self, address: List[int]) -> None:
        r"""Apply a SWAP gate on the local qubits with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to act on in the quantum register
        """
        self.__add_double_qubit_gate('cx', address)
        self.__add_double_qubit_gate('cx', [address[1], address[0]])
        self.__add_double_qubit_gate('cx', address)

    def measure(self, address: int, basis: Optional[str] = "z") -> None:
        r"""Measure the qubit with a given address in the computational basis.

        Args:
            address (int): local address of the qubit to act on in the quantum register
            basis (str): on which basis should the measurement performs, either Z-basis or X-basis
        """
        if basis.casefold() == "z":
            self.__add_single_qubit_gate("measure", address)
        elif basis.casefold() == "x":
            self.h(address)
            self.__add_single_qubit_gate("measure", address)

        # Side effect for measurement, the trick here is to use the qubit register as its measurement outcome
        which_qubit = int(self.units[address]['qubit'].state[0][0])
        self.units[address]['outcome'] = which_qubit

        # Rest the unit once measured
        self.reset(address)

    def bsm(self, address: List[int]) -> None:
        r"""Perform Bell state measurement on the qubits with given addresses.

        Args:
            address (List[int]): local addresses of the qubits to perform Bell state measurement in the quantum register
        """
        self.cnot([address[0], address[1]])
        self.h(address[0])
        self.measure(address[0])
        self.measure(address[1])



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for QKD node templates with different functionalities.
"""

from typing import List
from qcompute_qnet.core.des import Entity, EventHandler
from qcompute_qnet.topology.node import Node, Satellite
from qcompute_qnet.protocols.protocol import ProtocolStack
from qcompute_qnet.devices.source import PhotonSource
from qcompute_qnet.devices.detector import PolarizationDetector
from qcompute_qnet.models.qkd.application import QKDApp, QKDAppL4
from qcompute_qnet.models.qkd.resource_management import QKDRMP
from qcompute_qnet.models.qkd.routing import QKDRouting, QKDRoutingL2
from qcompute_qnet.models.qkd.key_generation import BB84, DecoyBB84, KeyGeneration
from qcompute_qnet.models.qkd.message import QKDMessage

__all__ = [
    "QKDNode",
    "EndNode",
    "TrustedRepeaterNode",
    "BackboneNode",
    "QKDSatellite",
    "RMPEndNode",
    "RMPRepeaterNode"
]


class QKDNode(Node):
    r"""Class for the simulation of a QKD node.

    Attributes:
        location (Tuple): geographical location of the QKD node
        photon_source (PhotonSource): photon source of the QKD node
        polar_detector (PolarizationDetector): polarization detector of the QKD node
        keys (Dict[Node, List]): generated keys with other QKD nodes
    """

    def __init__(self, name: str, env=None, location=None):
        r"""Constructor for QKDNode class.

        Args:
            name (str): name of the QKD node
            env (DESEnv): related discrete-event simulation environment
            location (Tuple): geographical location of the QKD node
        """
        super().__init__(name, env)
        self.location = location
        self.photon_source = PhotonSource("photon_source")
        self.polar_detector = PolarizationDetector("polar_detector")
        self.install([self.photon_source, self.polar_detector])
        ps = ProtocolStack(f"ps_{self.name}")
        self.load_protocol(ps)
        self.keys = {}

    def set_key_generation(self, peer=None, **kwargs) -> "KeyGeneration":
        r"""Set a key generation protocol with given parameters.

        Args:
            peer (Node): peer node of the key generation protocol
            **kwargs: keyword arguments to set

        Returns:
            KeyGeneration: key generation protocol instance created

        Note:
            This method will set the peer node and other related parameters of the key generation protocol of a
            ``QKDNode`` and return a protocol instance.

        Hint:
            If the parameters are not specified, default parameters are used in the key generation protocol.

        Examples:
            It is a universal approach to parse the parameters as a dictionary.

            >>> alice = QKDNode("Alice")
            >>> bob = QKDNode("Bob")
            >>> intensities = {"prob": [0.5, 0.25, 0.25], "mean_photon_num": [0.8, 0.1, 0]}
            >>> tx_options = {"protocol": "DecoyBB84", "tx_bases_ratio": [0.8, 0.2], "intensities": intensities}
            >>> rx_options = {"protocol": "DecoyBB84", "rx_bases_ratio": [0.6, 0.4]}
            >>> bb84_alice = alice.set_key_generation(peer=bob, **tx_options)
            >>> bb84_bob = bob.set_key_generation(peer=alice, **rx_options)
        """
        # If the parameters are specified, reset related configurations in the protocol stack
        if kwargs != {}:
            for attr in kwargs:
                if attr == "protocol":
                    assert isinstance(kwargs[attr], str), "'protocol' should be a str value."
                elif attr == "tx_bases_ratio":
                    assert isinstance(kwargs[attr], list), "'tx_bases_ratio' should be a list."
                elif attr == "rx_bases_ratio":
                    assert isinstance(kwargs[attr], list), "'rx_bases_ratio' should be a list."
                elif attr == "intensities":
                    assert isinstance(kwargs[attr], dict), "'intensities' should be a dict."
                    assert "prob" in kwargs[attr].keys() and "mean_photon_num" in kwargs[attr].keys(), \
                        "'intensities' should include 'prob' and 'mean_photon_num'."
                    assert isinstance(kwargs[attr]["prob"], list), "'prob' should be a list"
                    assert isinstance(kwargs[attr]["mean_photon_num"], list), "'mean_photon_num' should be a list"
                else:
                    raise TypeError(f"Setting {attr} is not allowed.")
            self.protocol_stack.set(protocol_type="key_generation", **kwargs)

        # Else, set with default parameters if not configure before
        elif kwargs == {} and "key_generation" not in self.protocol_stack.config.keys():
            print(f"Note: Default parameters are set for {self.name}'s key generation protocol.")
            self.protocol_stack.set(protocol_type="key_generation", protocol="BB84")

        if peer is not None:
            if self.protocol_stack.config['key_generation']['protocol'].casefold() == "BB84".casefold():
                key_gen_proto = BB84(f"BB84_{self.name}_{peer.name}", peer)
                if 'tx_bases_ratio' in self.protocol_stack.config['key_generation'].keys():
                    key_gen_proto.set(tx_bases_ratio=self.protocol_stack.config['key_generation']['tx_bases_ratio'])
                if 'rx_bases_ratio' in self.protocol_stack.config['key_generation'].keys():
                    key_gen_proto.set(rx_bases_ratio=self.protocol_stack.config['key_generation']['rx_bases_ratio'])

            elif self.protocol_stack.config['key_generation']['protocol'].casefold() == "DecoyBB84".casefold():
                key_gen_proto = DecoyBB84(f"DecoyBB84_{self.name}_{peer.name}", peer)
                if 'tx_bases_ratio' in self.protocol_stack.config['key_generation'].keys():
                    key_gen_proto.set(tx_bases_ratio=self.protocol_stack.config['key_generation']['tx_bases_ratio'])
                if 'rx_bases_ratio' in self.protocol_stack.config['key_generation'].keys():
                    key_gen_proto.set(rx_bases_ratio=self.protocol_stack.config['key_generation']['rx_bases_ratio'])
                if 'intensities' in self.protocol_stack.config['key_generation'].keys():
                    key_gen_proto.set_intensity(self.protocol_stack.config['key_generation']['intensities'])

            else:
                raise TypeError(f"Protocol {self.protocol_stack.config['key_generation']['protocol']}"
                                f" is not supported in this version.")

            return key_gen_proto

    def receive_classical_msg(self, src: "Node", msg: "ClassicalMessage") -> None:
        r"""Receive a classical message from the classical channel.

        The source node is inferred by the receiver and is not part of the transmitted message.

        Args:
            src (Node): source of the classical message
            msg (ClassicalMessage): classical message to receive
        """
        if msg.dst == self:
            for proto in self.protocol_stack.protocols:
                if isinstance(proto, msg.protocol):
                    proto.receive_classical_msg(msg)
        else:  # find the routing protocol to deal with the message
            for proto in self.protocol_stack.protocols:
                if isinstance(proto, QKDRouting):
                    proto.receive_classical_msg(msg)

    def receive_quantum_msg(self, src: "Node", msg: "QuantumMessage") -> None:
        r"""Receive a quantum message from the quantum channel.

        Args:
            src (Node): sender node of the quantum message
            msg (QuantumMessage): quantum message to receive
        """
        for proto in self.protocol_stack.protocols:
            if isinstance(proto, KeyGeneration):
                proto.receive_quantum_msg(msg)


class EndNode(QKDNode):
    r"""Class for the simulation of an end node.

    Note that ``EndNode`` usually represents a user node.

    Attributes:
        direct_repeater (TrustedRepeaterNode): directly connected repeater
    """

    def __init__(self, name: str, env=None, location=None, rmp: bool = False):
        r"""Constructor for EndNode class.

        Args:
            name (str): name of the end node
            env (DESEnv): related discrete-event simulation environment
            location (Tuple): geographical location of the end node
            rmp (bool): whether to adopt the resource management protocol for quantum key distribution
        """
        super().__init__(name, env, location)
        self.direct_repeater = None
        routing = QKDRoutingL2(f"QKDRoutingL2_{self.name}") if rmp is True else QKDRouting(f"QKDRouting_{self.name}")
        self.protocol_stack.build(routing)
        self.load_protocol(self.protocol_stack)

    def init(self) -> None:
        r"""Node initialization and assign the direct repeater if there is one.
        """
        super().init()
        for node in self.links.keys():
            if isinstance(node, TrustedRepeaterNode):
                self.direct_repeater = node

    @property
    def direct_nodes(self) -> List:
        r"""Return nodes that directly connect with the end node.

        Returns:
            List: a list of nodes that are directly connected with the node
        """
        return list(self.links.keys())

    def key_request(self, **kwargs) -> None:
        r"""Generate a QKD request.

        Args:
            **kwargs: keyword arguments for a key request
        """
        qkd_app = self.set_qkd_app()
        qkd_app.start(**kwargs)

    def set_qkd_app(self) -> "QKDApp":
        r"""Set QKDApp protocol and update the protocol stack.

        Returns:
            QKDApp: QKDApp protocol instance created
        """
        for proto in self.protocol_stack.protocols:
            # TODO: only consider QKDRouting as a lower protocol in this version
            if isinstance(proto, QKDRouting):
                qkd_routing = proto
        qkd_app = QKDApp(f"QKDApp_{self.name}")
        self.protocol_stack.update(qkd_app, lower_protocols=[qkd_routing])

        return qkd_app

    def set_key_generation(self, peer=None, **kwargs) -> "KeyGeneration":
        r"""Set a key generation protocol with given parameters.

        Args:
            peer (Node): peer node of the key generation protocol
            **kwargs: keyword arguments to set

        Returns:
            KeyGeneration: key generation protocol instance created
        """
        key_gen_proto = super().set_key_generation(peer, **kwargs)

        if key_gen_proto is not None:
            for proto in self.protocol_stack.protocols:
                # TODO: only consider QKDRouting as a lower protocol in this version
                if isinstance(proto, QKDRouting):
                    qkd_routing = proto
            self.protocol_stack.update(key_gen_proto, upper_protocols=[qkd_routing])

            return key_gen_proto


class TrustedRepeaterNode(QKDNode):
    r"""Class for the simulation of a trusted repeater node.

    Attributes:
        classical_routing_table (Dict[Node, Node]): classical routing table the trusted repeater holds
        quantum_routing_table (Dict[Node, Node]): quantum routing table the trusted repeater holds
    """

    def __init__(self, name: str, env=None, location=None, rmp: bool = False):
        r"""Constructor for TrustedRepeaterNode class.

        Args:
            name (str): name of the trusted repeater
            env (DESEnv): related discrete-event simulation environment
            location (Tuple): geographical location of the trusted repeater
            rmp (bool): whether to adopt the resource management protocol for quantum key distribution
        """
        super().__init__(name, env, location)
        self.classical_routing_table = {}
        self.quantum_routing_table = {}
        routing = QKDRoutingL2(f"QKDRoutingL2_{self.name}") if rmp is True else QKDRouting(f"QKDRouting_{self.name}")
        self.protocol_stack.build(routing)
        self.load_protocol(self.protocol_stack)

    def set_key_generation(self, peer=None, **kwargs) -> "KeyGeneration":
        r"""Set a key generation protocol with given parameters.

        Args:
            peer (Node): peer node of the key generation protocol
            **kwargs: keyword arguments to set

        Returns:
            KeyGeneration: key generation protocol instance created
        """
        key_gen_proto = super().set_key_generation(peer, **kwargs)

        if key_gen_proto is not None:
            for proto in self.protocol_stack.protocols:
                # TODO: only consider QKDRouting as an upper protocol in this version
                if isinstance(proto, QKDRouting):
                    qkd_routing = proto
            self.protocol_stack.update(key_gen_proto, upper_protocols=[qkd_routing])

            return key_gen_proto


class BackboneNode(TrustedRepeaterNode):
    r"""Class for the simulation of a backbone node.
    """

    def __init__(self, name: str, env=None, location=None):
        r"""Constructor for BackboneNode class.

        Args:
            name (str): name of the backbone node
            env (DESEnv): related discrete-event simulation environment
            location (Tuple): geographical location of the backbone node
        """
        super().__init__(name, env, location)


class QKDSatellite(QKDNode, Satellite):
    r"""Class for satellites capable of implementing quantum key distribution.
    """

    def __init__(self, name: str, env=None, location=None):
        r"""Constructor for QKDSatellite class.

        Args:
            name (str): name of the QKD satellite node
            env (DESEnv): related discrete-event simulation environment
            location (Tuple): geographical location of the QKD satellite node
        """
        super().__init__(name, env, location)


class RMPEndNode(EndNode):
    r"""Class for the simulation of an end node in resource management architecture.

    Note that ``RMPEndNode`` usually represents a user node.

    Attributes:
        idle (bool): whether the node is idle
        random_request (bool): whether to generate random requests automatically
        reqs_delivered (int): number of the delivered requests
        reqs_discarded (int): number of the discarded requests
        keys_delivered (int): number of the delivered keys
        keys_discarded (int): number of the discarded keys
    """

    def __init__(self, name: str, env=None, location=None):
        r"""Constructor for RMPEndNode class.

        Args:
            name (str): name of the RMPEndNode
            env (DESEnv): related discrete-event simulation environment
            location (Tuple): geographical location of the RMPEndNode
        """
        super().__init__(name, env, location, rmp=True)
        self.idle = True
        self.random_request = False
        self.reqs_delivered = 0
        self.reqs_discarded = 0
        self.keys_delivered = 0
        self.keys_discarded = 0

        rmp = QKDRMP(f"QKDRMP_{self.name}", self)
        for proto in self.protocol_stack.protocols:
            if isinstance(proto, QKDRoutingL2):
                qkd_routing = proto
        self.protocol_stack.update(rmp, lower_protocols=[qkd_routing])

    def init(self) -> None:
        r"""RMPEndNode initialization.
        """
        super().init()
        if self.random_request:
            # Generate a random request
            self.random_req()

    def random_req(self) -> None:
        r"""Random key request generation for the demo of Beijing quantum metropolitan area network.
        """
        import random
        import numpy as np

        if self.reqs_delivered == self.reqs_discarded == 0:
            # Start time for the first request is sampled from a Gaussian distribution
            start_time = int(abs(random.gauss(1e10, 5e9)))
        else:
            # Start time for other requests is sampled from a Poisson distribution
            start_time = int(abs(np.random.poisson(1e7, 1)) * 2e3)

        # Requested key number and key length
        key_num = 10
        key_length = 32

        # Search for other idle RMPEndNodes
        end_list = [node for node in self.owner.nodes
                    if isinstance(node, RMPEndNode) and node != self and node.idle is True]

        if len(end_list) != 0:
            # Shuffle the list of end nodes, and randomly choose one peer node for QKD request
            r = random.Random(start_time)
            r.shuffle(end_list)
            dst_node = end_list.pop()

            # Generate a QKD request
            handler = EventHandler(self, "key_request", dst=dst_node, key_num=key_num, key_length=key_length)
            self.scheduler.schedule_after(start_time, handler)
        else:
            return

    def key_request(self, **kwargs) -> None:
        r"""Generate a QKD request.

        Args:
            **kwargs (Any): keyword arguments for a key request
        """
        if self.idle and kwargs['dst'].idle:
            if kwargs['dst'].random_request is False and self.random_request is True:
                self.random_req()
                return
            # Set the status of the sender and receiver to be busy
            self.set_busy()
            kwargs['dst'].set_busy()
            qkd_app = self.set_qkd_app()
            qkd_app.start(**kwargs)
        else:
            # Set a random request for another time point
            self.random_req()

    def set_qkd_app(self) -> "QKDAppL4":
        r"""Set QKDAppL4 protocol and update the protocol stack.

        Returns:
            QKDAppL4: the QKDAppL4 protocol instance created
        """
        for proto in self.protocol_stack.protocols:
            if isinstance(proto, QKDRMP):
                qkd_rmp = proto
        qkd_app = QKDAppL4(f"QKDApp_{self.name}")
        self.protocol_stack.update(qkd_app, lower_protocols=[qkd_rmp])

        return qkd_app

    def set_busy(self) -> None:
        r"""Set the end node to be busy.
        """
        self.idle = False

    def set_idle(self) -> None:
        r"""Set the end node to be idle.
        """
        self.idle = True


class RMPRepeaterNode(TrustedRepeaterNode):
    r"""Class for the simulation of a repeater node with resource management.

    Notes:
        For each connected link of a ``RMPRepeaterNode``, there is a pair of photon source and polarization detector
        for quantum key distribution.

    Attributes:
        key_pools (Dict): a list of the key pools owned by the node
        reqs_delivered (int): number of delivered requests
        keys_delivered (int): number of rejected requests
        reqs_discarded (int): number of discarded requests
        keys_discarded (int): number of discarded keys
        reqs_rejected (int): number of rejected requests
        keys_rejected (int): number of rejected keys
        reqs_invalid (int): number of invalid requests
        keys_invalid (int): number of invalid keys
        key_pool_volume (int): maximum volume of the key pools
        key_length (int): length of the keys stored in the key pools
    """

    def __init__(self, name: str, env=None, location=None, key_pool_volume: int = 20, key_length: int = 32):
        r"""Constructor of the RMPRepeaterNode class.

        Args:
            name (str): name of the RMPRepeaterNode
            env (DESEnv): related discrete-event simulation environment
            location (Tuple): geographical location of the RMPRepeaterNode
            key_pool_volume (int): maximum volume of the key pools
            key_length (int): length of the keys stored in the key pools
        """
        super().__init__(name, env, location, rmp=True)
        self.key_pools = {}

        self.reqs_delivered = 0
        self.reqs_discarded = 0
        self.reqs_invalid = 0
        self.reqs_rejected = 0

        self.keys_delivered = 0
        self.keys_discarded = 0
        self.keys_invalid = 0
        self.keys_rejected = 0

        self.key_pool_volume = key_pool_volume
        self.key_length = key_length

        rmp = QKDRMP(f"QKDRMP_{self.name}", self)
        self.protocol_stack.build(rmp)
        for proto in self.protocol_stack.protocols:
            if isinstance(proto, QKDRoutingL2):
                qkd_routing = proto
        self.protocol_stack.update(rmp, lower_protocols=[qkd_routing])

    def init(self) -> None:
        r"""RMPRepeaterNode initialization.
        """
        self.init_photon_sources()
        self.init_polar_detectors()
        self.init_key_pools(max_volume=self.key_pool_volume, key_length=self.key_length)

    def init_photon_sources(self) -> None:
        r"""Initialization of the photon sources. One photon source is set for each connected link.
        """
        photon_sources = {}
        for neighbor_node in self.links:
            photon_source = PhotonSource(f"photon_source_to_{neighbor_node.name}")
            self.install(photon_source)
            photon_sources[neighbor_node] = photon_source

        self.photon_source = photon_sources

    def init_polar_detectors(self) -> None:
        r"""Initialization of the polar detectors. One polar detector is set for each connected link.
        """
        polar_detectors = {}
        for neighbor_node in self.links:
            polar_detector = PolarizationDetector(f"polar_detector_from_{neighbor_node.name}")
            self.install(polar_detector)
            polar_detectors[neighbor_node] = polar_detector

        self.polar_detector = polar_detectors

    def init_key_pools(self, max_volume: int, key_length: int) -> None:
        r"""Initialization of the key pools with other RMPRepeaterNodes.

        Args:
            max_volume (int): maximum volume of the key pools
            key_length (int): length of keys of the key pools
        """
        for neighbor_node in self.links:
            if isinstance(neighbor_node, RMPRepeaterNode):
                # Set two key pools for an upstream role and a downstream role in key generation, respectively
                upstream_key_pool = self.KeyPool(node=self,
                                                 peer_node=neighbor_node,
                                                 key_length=key_length,
                                                 max_volume=max_volume)
                downstream_key_pool = self.KeyPool(node=self,
                                                   peer_node=neighbor_node,
                                                   key_length=key_length,
                                                   max_volume=max_volume)
                self.key_pools[neighbor_node] = [upstream_key_pool, downstream_key_pool]
                self.install([upstream_key_pool, downstream_key_pool])

                # Start key generation
                handler_up = EventHandler(upstream_key_pool, "key_gen", peer=neighbor_node, towards='up')
                handler_down = EventHandler(downstream_key_pool, "key_gen", peer=neighbor_node, towards='down')
                self.scheduler.schedule_now(handler_up)
                self.scheduler.schedule_now(handler_down)

    def set_key_generation(self, peer=None, **kwargs) -> "KeyGeneration":
        r"""Set a key generation protocol with given parameters.

        Args:
            peer (Node): peer node of the key generation protocol
            **kwargs (Any): keyword arguments to set

        Returns:
            KeyGeneration: key generation protocol instance created
        """
        from qcompute_qnet.models.qkd.key_generation import PrepareAndMeasure

        towards = kwargs.pop('towards') if kwargs.get('towards') else None
        role = PrepareAndMeasure.Role.TRANSMITTER if towards == 'up' else PrepareAndMeasure.Role.RECEIVER

        for proto in self.protocol_stack.protocols:
            if isinstance(proto, QKDRoutingL2):
                qkd_routing_l2 = proto

        # Reuse the existed key generation instance
        for proto in qkd_routing_l2.lower_protocols:
            if proto.peer == peer and proto.role == role:
                self.env.logger.info(f"{self.name} found an existed {proto.name}.{proto.role},"
                                     f"reused it as the key generation protocol")
                proto.role = None
                return proto

        key_gen_proto = super().set_key_generation(peer, **kwargs)
        return key_gen_proto

    def receive_quantum_msg(self, src: "Node", msg: "QuantumMessage") -> None:
        r"""Receive a quantum message from the quantum channel.

        Args:
            src (Node): sender node of the quantum message
            msg (QuantumMessage): quantum message to receive
        """
        for proto in self.protocol_stack.protocols:
            if isinstance(proto, KeyGeneration):
                proto.receive_quantum_msg(msg, src=src)

    class KeyPool(Entity):
        r"""Class for the structure that stores the sifted keys between repeaters.

        Attributes:
            node (Node): owner of the key pool
            peer_node (Node): peer node for key generation
            key_length (int): length of the keys in the key pool
            max_volume (int): maximum number of keys the key pool can accommodate
            current_volume (int): number of available keys
            regenerate_threshold (int): minimum number of keys needed to be held for providing service
            keys (List[str]): list of the sifted keys
            idle (bool): whether the key pool is idle
            ready_to_generate (bool): whether the key pool is ready for key generation
            generating (bool): whether the key pool is generating keys
        """

        def __init__(self, node: "Node", peer_node: "Node", key_length: int, max_volume: int,
                     name: str = None, env=None):
            r"""Constructor of KeyPool class.

            Args:
                node (Node): owner of the key pool
                peer_node (Node): peer repeater node for key generation
                key_length (int): length of the keys in the key pool
                max_volume (int): maximum number of keys the key pool can accommodate
                name (str): name of the key pool
                env (DESEnv): simulation environment of the key pool
            """
            super().__init__(name, env)
            self.node = node
            self.peer_node = peer_node

            self.key_length = key_length
            self.max_volume = max_volume
            self.regenerate_threshold = max_volume // 4
            self.current_volume = 0

            self.keys = []
            self.idle = True
            self.ready_to_generate = True
            self.generating = False

        def init(self) -> None:
            r"""KeyPool initialization.
            """
            self.owner = self

        def key_gen(self, peer: "Node", towards: str, key_num: int = None) -> None:
            r"""Key generation function for the key pool.

            Args:
                peer (Node): the node sharing the keys
                towards (str): denote the direction of the key pool
                key_num (int): the number of sifted keys to be generated
            """
            self.node.env.logger.info(f"{self.node.name}'s {towards}stream key pool"
                                      f" started key generation with {peer.name}")

            # Set the status of the key pool
            self.idle = False
            self.generating = True
            self.ready_to_generate = False

            key_num = key_num if key_num is not None else self.max_volume
            # Create a 'GENERATE' message to inform QKDRoutingL2 to generate keys
            generate_msg = QKDMessage(src=self.node,
                                      dst=peer,
                                      protocol=QKDRoutingL2,
                                      data={'type': QKDMessage.Type.GENERATE,
                                            'key_num': key_num,
                                            'key_length': self.key_length})

            for proto in self.node.protocol_stack.protocols:
                if isinstance(proto, QKDRoutingL2):
                    qkd_routing_l2 = proto
                    # Schedule an event for QKDRoutingL2 to generate keys
                    scheduler = EventHandler(qkd_routing_l2, "receive_classical_msg", msg=generate_msg, towards=towards)
                    self.node.scheduler.schedule_now(scheduler)
                    break

        def fulfill(self, request: "QKDMessage") -> bool:
            r"""Judge if a request can be fulfilled with current volume of the key pool.

            Args:
                request (QKDMessage): request to execute

            Returns:
                bool: whether the request can be fulfilled
            """
            res = 0
            if request.data['type'] == QKDMessage.Type.PATH:
                res = self.max_volume - request.data['key_num']
            elif request.data['type'] == QKDMessage.Type.RESERVE:
                res = self.current_volume - request.data['key_num']

            return True if res > 0 else False

        def allocate(self, key_num: int, key_length: int, **kwargs) -> List:
            r"""Allocate keys to a request.

            Args:
                key_num (int): number of required keys
                key_length (int): length of required keys
                **kwargs: keyword arguments to check the volume and generate new keys

            Returns:
                List: a list of keys allocated for the request
            """
            # Calculate the number of keys to be allocated
            allocate_num = key_num * key_length // self.key_length
            assert self.current_volume > allocate_num, "No enough keys to allocate, should be already filtered."

            # Update current volume
            self.current_volume -= allocate_num
            # Allocate the keys
            allocate_keys = self.keys[0: allocate_num]
            self.keys = self.keys[allocate_num:]
            # Check if the key pools needs to start key generation
            towards = kwargs['towards']
            self.check_volume(towards)

            return allocate_keys

        def check_volume(self, towards: str) -> None:
            r"""Check the volume of the key pool to decide whether to start key generation.

            Args:
                towards (str): denote the direction of the key pool
            """
            if self.current_volume <= self.regenerate_threshold:
                # Prepare to start key generation for the key pool
                key_num = self.max_volume - self.current_volume
                # Reverse the towards for later operations
                towards = 'down' if towards == 'up' else 'up'
                dual_key_pool_index = 1 if towards == 'up' else 0
                dual_key_pool = self.peer_node.key_pools[self.node][dual_key_pool_index]

                self.ready_to_generate = True

                if dual_key_pool.ready_to_generate is True:
                    # Set a delay for the key generation
                    delay = self.node.links[self.peer_node].components[-1].channel2_1.delay * 5
                    handler = EventHandler(self, "key_gen", [self.peer_node, towards, key_num])
                    self.node.scheduler.schedule_after(delay, handler, None)
                    # Reverse direction for the peer node's key generation
                    towards = 'down' if towards == 'up' else 'up'
                    dual_handler = EventHandler(dual_key_pool, "key_gen", [self.node, towards, key_num])
                    self.peer_node.scheduler.schedule_after(delay, dual_handler, None)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for utility functions in quantum package.
"""

from argparse import ArgumentTypeError
from typing import List, Union
import numpy

__all__ = [
    "kron",
    "print_progress"
]


def kron(matrices: List[numpy.ndarray]) -> numpy.ndarray:
    r"""Take the kronecker product of a list of matrices.

    .. math::

        [A, B, C, \cdots] \to A \otimes B \otimes C \otimes \cdots

    Args:
        matrices (List[numpy.ndarray]): a list of matrix to product

    Returns:
        numpy.ndarray: the kronecker result
    """
    if not isinstance(matrices, list):
        raise ArgumentTypeError(f"Input {matrices} should be a list.")

    result = matrices[0]
    if len(matrices) > 1:  # kron together
        for i in range(1, len(matrices)):
            result = numpy.kron(result, matrices[i])
    return result


def print_progress(current_progress: Union[float, int], progress_name, track=True):
    r"""Print a progress bar.

    Args:
        current_progress (Union[float, int]): current progress percentage
        progress_name (str): name of the progress bar
        track (bool): whether to print the progress on the terminal
    """
    if current_progress < 0 or current_progress > 1:
        raise ArgumentTypeError(f"Invalid current progress: ({current_progress})!\n"
                                f"'current_progress' must be a percentage between 0 and 1")
    if not isinstance(track, bool):
        raise ArgumentTypeError(f"Invalid parameter ({track}) with the type `{type(track)}`!\n"
                                f"Only `bool` is supported as the parameter.")
    if track:
        print("\r" + f"{progress_name.ljust(30)}"
                     f"|{'■' * int(50 * current_progress):{50}s}| "
                     f"\033[94m {'{:6.2f}'.format(100 * current_progress)}% \033[0m ", flush=True, end="")
        if current_progress == 1:
            print(" (Done)")
